/*
 * civil_datastruct.c
 *
 *  Created on: Jul 14, 2009
 *      Author: yanming
 */

#define MAX_NAME_LEN 100
#define MAX_FILE_NAME_LEN 1000
#define MAX_LEN 20000                   /* The maximum possible line length of input files (parameter files, data files, etc.) */

typedef char name_t[MAX_NAME_LEN]; /**< Type for individual's name.*/
typedef char file_name_t[MAX_FILE_NAME_LEN]; /**< Type for file names.*/

/**
 * Data type for inheritance vectors.
 *
 * Each bit of a variable of this type corresponds to one meiosis indicator.
 * The length of the inheritance vector is limited to 30.
 */
typedef unsigned long iv_t;

/**
 * Data type for founder allelic types.
 *
 * Each bit of a variable of this type corresponds to allelic type of one
 * founder. Only two allelic types are possible and total number of founders
 * is limited to 30. In fact, when the total number of founders in a pedigree
 * exceeds 20, it becomes infeasible to compute a w-score for such a pedigree.
 */
typedef unsigned long fat_t;

typedef struct
{
   int   id;
   name_t name;
} name_id_t;   /**< (id, name) pair. Used mainly for matching names with ids. */

/**
 * Data type for an individual in a pedigree.
 */
typedef struct
{
   int   id; /**< id */
   int   pa_id;/**< father's id */
   int   ma_id;/**< mother's id */
   name_t name;  /**< name of the individual */
   name_t pa_name;  /**<father's name */
   name_t ma_name;  /**<mother's name */
   int   cmp;/**< component number */
   int   pa_mi;/**< index to paternal meiosis indicator */
   int   ma_mi;/**< index to maternal meiosis indicator */

   int   obsd;/**< whether trait value is observed for this individual */
   double y;  /**< observed trait value. Undefined, if not observed. */
} ind_t;

/**
 * Data type for a pedigree. A pedigree is a list of
 * individuals with parents specified.
 */
typedef struct
{
   char  name[MAX_NAME_LEN]; /**< name of the pedigree */
   int   size;/**< size (number of individuals) in the pedigree */
   ind_t *inds;  /**< the list of individuals */
} ped_t;

/**
 * Data type for a list of pedigrees.
 */
typedef struct
{
   int   n_peds;/**< number of pedigrees in the list */
   ped_t *peds;  /**< the pedigrees */
   double mu;  /**< sample mean of the trait values */
   double sd;  /**< sample sd of the trait values */
} ped_list_t;

/**
 * Allele descent graph. An alternative way to summarize a pedigree. Treat the
 * pedigree as a collection of alleles.
 *
 */
typedef struct
{
   int   id;/**< pedigree id */
   char  name[MAX_NAME_LEN]; /**< name of the pedigree */

   int   n_a;/**< number of alleles; */
   int   n_fa;/**< number of founder alleles; */
   int  *i_fa; /**< indices to founder alleles; */
   int   n_nfa;/**< number of nonfounder alleles; */
   int  *i_nfa; /**< indices to nonfounder alleles; */
   int  *i_maa; /**< indices to nonfounder alleles' parent's maternal allele; */
   int  *i_paa; /**< indices to nonfoudner alleles' parent's paternal allele; */

    /**int n_meis = n_nfa; */
   int  *i_mi; /**< indices to corresponding meiosis indicators */
   int   n_fixed_bits;
   iv_t *fixed_bit_masks;
   iv_t *mei_grp_masks;
   int   n_non_fixed_bits;
   iv_t *non_fixed_bit_masks;

/* iv_t *red_s; ///< a table that maps full inheritance vectors to reduced inheritance vectors */
/* When the number of pedigrees is large, it is too expensive to store all */
/* the maps in memory. */

   int   n_obs;/**< number of observed individuals */
   int  *i_obsa; /**< indices to alleles of the observed individuals */
   double *y;  /**< trait values of observed individuals */
} adg_t;

/**
 * A list of allele descent graphs. One for each pedigree.
 */
typedef struct
{
   int   n_peds;/**< number of pedigrees in the pedigree. */
   adg_t *peds;  /**< the pedigrees as allele descent graphs. */
   double mu;  /**< sample mean of the trait values */
   double sd;  /**< sample sd of the trait values */
} adg_list_t;

/**
 * A sample of meiosis indicators matrices.
 */
typedef struct
{
   int   smpl_size;/**<sample size. */
   int   n_meis;/**< number of meioses. */
   int   n_loci;/**< number of loci. */
   double *map;  /**< locus map */
   char ***s;  /**< A smpl_size x n_meis x n_loci binary matrix. */
} mi_mat_t;

/**
 * A collection of inheritance vectors. smpl_size x n_loci x n_peds inheritance
 * vectors.
 */
typedef struct
{
   int   smpl_size;/**< sample size. */
   int   n_loci;/**< number of loci. */
   int   n_peds;/**< number of pedigrees. */
   double *map;  /**< locus map */
    /**int *n_meis; ///< numbers of meioses. */
   iv_t ***s;  /**< inheritance vectors (smpl_size x n_loci x n_peds). */
} iv_list_t;

/**
 * w-score table.
 */
typedef struct
{
    /** int n_peds; ///< number of pedigrees. */
   long  n_ws; /**< table sizes. */
   double *lgw;  /**< store the w-score tables */
} lgw_table_t;

/**
 * w-score tables.
 */
typedef struct
{
   int   n_peds;/**< number of pedigrees. */
   long *n_ws;  /**< table sizes. */
   double **lgw;  /**< store the w-score tables */
} lgw_tables_t;

/**
 * variance component score table.
 */
typedef struct
{
   long  n_rep_s;
   int   nz;
   double sumz;
   double zz;
   int  *sumphi;
   double *phiz;
   int  *phiphi;
} vc_table_t;

/**
 * Output directory and file names.
 */
typedef struct
{
   file_name_t dir_in, dir_out;/**< output directory */
   file_name_t name_ped;/**< name of the data set, will be used to name the output files.*/
   file_name_t file_ped, file_trait, file_mis, file_mis_rearranged; /**< input files */
   file_name_t file_miv_t, file_miv_p, file_civ_t, file_civ_p;
                          /**< output files */
} file_names_t;

/**
 * Information on the meiosis indicator file dumped by gl_auto.
 */
typedef struct
{
   int   smpl_size;/**< number of MCMC runs */
   int   n_components;/**< number of pedigree components */
   int  *n_meis_ped;  /**< number of meioses in each pedigree */
   int   n_meis;/**< total number of meioses in all pedigrees, so it should be less than 32767 */
   int   n_loci;/**< total number of loci used when creating the oscor file */
   double *locus_map;  /**< locus map */
   int   n_pos;/**< number of test positions, including marginal, conditional test positions and conditioning positions.*/
   int  *pos; /**< test and conditioning positions, which are a subset of all loci */
   int   by_component;/**< whether the gl_auto output is by component: 0 for no, other value for yes */
} data_pars_t;

/**
 * Data structure for the trait model.
 */
typedef struct
{
   char *ca;      /**< The number of D alleles in each possible vector of founder allelic types. */
   double pD;      /**< minor allele frequency. */
   double loddD;      /**< loddD = log(pD/(1-pD)); */
   /* mean trait values for dd, Dd, DD genotypes. Not needed for computing w-scores. Will use it for simulating trait values. */
   /* double mug[3];  */
 /**< standard deviation of the trait values */
   double sd;      /**< sample standard deviation of the trait values */
   double mu;      /**< sample mean trait value or user-specified mean trait value */
   double se;      /**< standard deviation of the residual effect (polygenic + environment).
          Not known in practice. Will estimate it by sd or from additive and resid variances (user specify which way).
          */
} trait_model_t; /**< Trait model parameters */

/**
 * Marginal test settings, test statistic values and p-values.
 */
typedef struct
{
   int   smpl_size;/**< MCMC sample size of the inheritance vectors */
   int   n_mc;/**< Monte Carlo sample size for the null */
   int   n_pos;/**< number of test positions */
   int  *pos; /**< test positions */
   double *map;  /**< map positions of the test positions */
   double **t;  /**< test statistics (w-scores, vc scores etc. smpl_size x n_pos */
   double **p;  /**< p-values. smpl_size x n_pos */
   char  test[10]; /**< test statistic: w or vc */
} miv_pars_t;

/**
 * Conditional test settings, test statistic values and p-values.
 */
typedef struct
{
   int   smpl_size;/**< MCMC sample size of the inheritance vectors. */
   int   n_mc; /**< Monte Carlo sample size for the null */
   int   n_pos;/**< number of test positions */
   int  *pos; /**< test positions */
   int  *posl; /**< left conditioning positions */
   int  *posr; /**< right conditioning positions */
   double *map;  /**< map positions of the test positions */
   double *mapl;  /**< map positions of the left conditioning positions */
   double *mapr;  /**< map positions of the right conditioning positions */
   double **t;  /**< test statistics. smpl_size x n_pos */
   double **p;  /**< p-values. smpl_size x n_pos */
   char  test[10]; /**< test statistic: w or vc */
} civ_pars_t;

/**
 * Free memory allocated to an adg_t variable.
 *
 * @parme pt points to an adg_t variable.
 */
void free_adg (adg_t * pt)
{
   free (pt->i_fa);
   free (pt->i_nfa);
   free (pt->i_maa);
   free (pt->i_paa);
   free (pt->i_mi);
   free (pt->fixed_bit_masks);
   free (pt->mei_grp_masks);
   free (pt->non_fixed_bit_masks);
/**  free(pt->red_s); */
   free (pt->i_obsa);
   free (pt->y);
}

/**
 * Free memory allocated to an adg_list_t variable.
 *
 * @parme pt points to an adg_list_t variable.
 */
void free_adg_list (adg_list_t * pt)
{
   int   k;
   for (k = 0; k < pt->n_peds; k++)
   {
      free_adg (&pt->peds[k]);
   }
   free (pt->peds);
}

/**
 * Free memory allocated to a ped_t variable.
 *
 * @param ped_pt points to a ped_t variable.
 */
void free_pedi (ped_t * ped_pt)
{
   free (ped_pt->inds);
    /**free(ped); */
}

/**
 * Free memory allocated to a ped_list_t variable.
 *
 * @param ped_list_pt points to a ped_list_t variable.
 *
 * Note: the pointer ped_list_pt itself will not be freed.
 */
void free_ped_list (ped_list_t * ped_list_pt)
{
   int   k;
   for (k = 0; k < ped_list_pt->n_peds; k++)
   {
      free_pedi (&ped_list_pt->peds[k]);
   }
   free (ped_list_pt->peds);
}

/**
 * Free memory allocated to a mi_mat_t variable.
 *
 * @param s_pt points to the mi_mat_t variable.
 *
 * Note: the pointer s_pt itself will not be freed.
 */
void free_mi_mat (mi_mat_t * s_pt)
{
   int   h;/** index sample points */
   int   i;/** index meioses */
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (i = 0; i < s_pt->n_meis; i++)
      {
         free (s_pt->s[h][i]);
      }
      free (s_pt->s[h]);
   }
   free (s_pt->map);
   free (s_pt->s);
}

/**
 * Free memory allocated to a iv_list_t variable.
 *
 * @param s_pt points to the iv_list_t variable.
 *
 * Note: the pointer s_pt itself will not be freed.
 */
void free_iv_list (iv_list_t * s_pt)
{
   int   h;/**ndex sample points */
   int   j;/**ndex loci */
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < s_pt->n_loci; j++)
      {
         free (s_pt->s[h][j]);
      }
      free (s_pt->s[h]);
   }
   free (s_pt->s);
   free (s_pt->map);
}

/**
 * Free memory allocated to a lgw_tables_t variable.
 *
 * @param lgw_pt points to the lgw_tables_t variable.
 */
void free_lgw_tables (lgw_tables_t * lgw_pt)
{
   int   k;                             /* index pedigrees; */
   for (k = 0; k < lgw_pt->n_peds; k++)
   {
      free (lgw_pt->lgw[k]);
   }
   free (lgw_pt->lgw);
   free (lgw_pt->n_ws);
}

/**
 * Free memory allocated to a lgw_tables_t variable.
 *
 * @param lgw_pt points to the lgw_tables_t variable.
 *
 */
void free_lgw_table (lgw_table_t * lgw_pt)
{
   free (lgw_pt->lgw);
}

/**
 * Free memory allocated to a vc_table_t variable.
 *
 * @param pt points to the vc_table_t variable.
 *
 */
void free_vc_table (vc_table_t * pt)
{
   free (pt->sumphi);
   free (pt->phiz);
   free (pt->phiphi);
}

/**
 * Free memory allocated to a data_pars_t variable.
 *
 * @param pt points to a data_pars_t variable.
 *
 */
void free_data_pars (data_pars_t * pt)
{
   free (pt->locus_map);
   free (pt->pos);
   free (pt->n_meis_ped);
}

/**
 * Free memory allocated to a trait_model_t variable.
 *
 * @param pt points to a trait_model_t variable.
 *
 */
void free_trait_model (trait_model_t * pt)
{
   free (pt->ca);
}

/**
 * Free memory allocated to a miv_pars_t variable.
 *
 * @param pt points to a miv_pars_t variable.
 */
void free_miv_pars (miv_pars_t * pt)
{
   int   h;

   for (h = 0; h < pt->smpl_size; h++)
   {
      free (pt->t[h]);
      free (pt->p[h]);
   }
   free (pt->t);
   free (pt->p);
   free (pt->pos);
   free (pt->map);
}

/*
 * Free memory allocated to a miv_pars_t variable.
 *
 * @param pt points to a miv_pars_t variable.
 */
void free_civ_pars (civ_pars_t * pt)
{
   int   h;

   for (h = 0; h < pt->smpl_size; h++)
   {
      free (pt->t[h]);
      free (pt->p[h]);
   }
   free (pt->t);
   free (pt->p);
   free (pt->pos);
   free (pt->posl);
   free (pt->posr);
   free (pt->map);
   free (pt->mapl);
   free (pt->mapr);
}

/**
 * Print the bits in a binary representation of a integer. Lower bits first.
 *
 * @param s the integer
 *
 * @param n the length of the vector
 */
void print_bits (iv_t s, int n)
{
   int   i;
   for (i = 0; i < n; i++)
   {
      printf ("%2d", (int) ((s >> i) & 1));
   }
  /***rintf("\n"); */
}

/**
 * Print a pedigree.
 *
 * @param ped_pt pointer to a pedigree.
 */
void print_ped (ped_t * ped_pt)
{
   int   i;
   printf ("ped %s:\n", ped_pt->name);
   printf ("%5s %5s %5s %10s %10s %10s %3s %5s %5s %5s %6s\n",
           "id", "pa.id", "ma.id",
           "name", "pa.name", "ma.name",
           "cmp", "pa_mi", "ma_mi", "obsd", "trait");

   for (i = 0; i < ped_pt->size; i++)
   {
      /* printf( "%5d %5d %5d %10s %10s %10s %3d %5d %5d %5d\n", */
      printf ("%5d %5d %5d %10s %10s %10s %3d %5d %5d %5d %10.3f\n",
              ped_pt->inds[i].id, ped_pt->inds[i].pa_id, ped_pt->inds[i].ma_id,
              ped_pt->inds[i].name,
              ped_pt->inds[i].pa_name, ped_pt->inds[i].ma_name,
              ped_pt->inds[i].cmp, ped_pt->inds[i].pa_mi, ped_pt->inds[i].ma_mi,
              /*  ped_pt->inds[i].obsd); */
              ped_pt->inds[i].obsd, ped_pt->inds[i].y);
  /**ped_pt->inds[i].obsd, (ped_pt->inds[i].obsd)?ped_pt->inds[i].y:0); */
   }

   printf ("\n");
}

/**
 * Print meiosis indicator matrix.
 *
 * @param s_pt points to a meiosis indicator matrix.
 */
void print_mi_mat (mi_mat_t * s_pt)
{
   int   h;/**ndexes sample points */
   int   i;/**ndexes meioses */
   int   j;/**ndexes loci */

   printf ("Meiosis indicator matrix:\n");
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (i = 0; i < s_pt->n_meis; i++)
      {
         printf ("%4d: ", i);
         for (j = 0; j < s_pt->n_loci; j++)
         {
            printf ("%d", s_pt->s[h][i][j]);
         }
         printf ("\n");
      }
      printf ("\n");
   }
}

/**
 * Print an inheritance vector
 *
 * @param s the inheritance vector
 *
 * @param l_s the length of the inheritance vector
 */
void print_iv (iv_t s, int l_s)
{
   int   i;
   for (i = 0; i < l_s; i++)
   {
      printf ("%ld", (s >> i) & 1);
   }
    /**printf("\n"); */
}

/**
 * Print a list of inheritance vectatofs
 *
 * @param iv_list_pt points the list of inheritance vectors
 *
 * @param points the collection of pedigrees adg_list_pt. adg_list_pt will
 * provide information (such as number of meioses in each pedigree) needed to
 * convert a iv_t type inheritance vector to meiosis indicator bits.
 */
void print_iv_list (const iv_list_t * iv_list_pt,
                    const adg_list_t * adg_list_pt)
{
   int   h;/**indexes sample points */
   int   j;/**indexes loci */
   int   k;/**indexes pedigrees */
   /*   int i; // indexes meiosis indicators of observed individuals */

   printf ("Inheritance vectors:\n");
   for (h = 0; h < iv_list_pt->smpl_size; h++)
   {
      for (j = 0; j < iv_list_pt->n_loci; j++)
      {
         printf ("locus %3d: ", j);
         for (k = 0; k < iv_list_pt->n_peds; k++)
         {
            if (adg_list_pt == NULL)
            {
               printf ("%ld ", iv_list_pt->s[h][j][k]);
            }
            else
            {
               print_iv (iv_list_pt->s[h][j][k], adg_list_pt->peds[k].n_nfa);
               printf ("|");
            }
         }
         printf ("\n");
      }
      printf ("\n");
   }
}

/**
 * Print a list of inheritance vectors
 *
 * @param iv_list_pt points the list of inheritance vectors
 *
 * @param adg_list_pt points the collection of pedigrees. adg_list_pt will
 * provide information (such as number of meioses in each pedigree) needed to
 * convert a iv_t type inheritance vector to meiosis indicator bits.
 */
void print_iv_list2 (const iv_list_t * iv_list_pt,
                     const adg_list_t * adg_list_pt)
{
   int   h;                             /* indexes sample points */
   int   j;                             /* indexes loci */
   int   k;                             /* indexes pedigrees */
   int   i;                             /* indexes meiosis indicators of observed individuals */

   printf ("Inheritance vectors:\n");
   for (h = 0; h < iv_list_pt->smpl_size; h++)
   {
      printf ("Sample %4d\n", h);
      for (k = 0; k < iv_list_pt->n_peds; k++)
      {
         for (i = 0; i < adg_list_pt->peds[k].n_nfa; i++)
         {
            for (j = 0; j < iv_list_pt->n_loci; j++)
            {
               printf ("%ld", (iv_list_pt->s[h][j][k] >> i) & 1);
            }
            printf ("\n");
         }
         for (j = 0; j < iv_list_pt->n_loci; j++)
         {
            if (j % 10 == 0)
            {
               printf ("%d", j / 10);
            }
            else
            {
               printf ("+");
            }
         }
         printf ("\n");
      }
      printf ("\n");
   }
   printf ("\n");
}

/**
* Print a list of inheritance vectors
*
* @param iv_list_pt points the list of inheritance vectors
*
*/
void print_iv_list_3 (const iv_list_t * iv_list_pt)
{
   int   h;
        /**indexes sample points */
   int   j;
        /**indexes loci */
   int   k;
        /**indexes pedigrees */
 /**   int i; // indexes meiosis indicators of observed individuals */

   printf ("Inheritance vectors:\n");
   for (h = 0; h < iv_list_pt->smpl_size; h++)
   {
      for (k = 0; k < iv_list_pt->n_peds; k++)
      {
         for (j = 0; j < iv_list_pt->n_loci; j++)
         {
    /**           printf("locus %3d: ", j); */
    /**              if (adg_list_pt == NULL) { */
            printf ("%ld ", iv_list_pt->s[h][j][k]);
         }
   /**     printf("\n"); */
      }
      printf ("\n");
   }
}

/**
 * Print an allele descent graph.
 *
 * @param adg_pt pointer to a allele descent graph.
 */
void print_adg (const adg_t * adg_pt)
{
   int   i;/**ndexes alleles */

   printf
      ("===========================================================================\n");
   printf ("Pedigree name: %s\n", adg_pt->name);
   printf ("Number of alleles: %d\n", adg_pt->n_a);
   printf ("Number of founder alleles: %d\n", adg_pt->n_fa);
   printf ("Indices to founder alleles:");
   for (i = 0; i < adg_pt->n_fa; i++)
   {
      printf ("%2d ", adg_pt->i_fa[i]);
   }
   printf ("\n");

   printf ("Number of nonfounder alleles: %d\n", adg_pt->n_nfa);

   printf ("Indices to nonfounder alleles:\n");
   printf ("Indices to parent's paternal alleles of nonfounder alleles:\n");
   printf ("Indices to parent's maternal alleles of nonfounder alleles:\n");
   printf ("Meiosis indicator indices to nonfounder alleles:\n");

   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      printf ("%3d ", adg_pt->i_nfa[i]);
   }
   printf ("\n");

   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      printf ("%3d ", adg_pt->i_paa[i]);
   }
   printf ("\n");

   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      printf ("%3d ", adg_pt->i_maa[i]);
   }
   printf ("\n");

   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      printf ("%3d ", adg_pt->i_mi[i]);
   }
   printf ("\n\n");

   printf ("Meiosis group masks, fixed bit masks\n");
   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      printf ("%d", adg_pt->i_nfa[i] % 10);
   }
   printf ("\n");

    /**printf("%d %d \n", adg_pt->n_fixed_bits, adg_pt->n_non_fixed_bits); */
   for (i = 0; i < adg_pt->n_fixed_bits; i++)
   {
      print_iv (adg_pt->mei_grp_masks[i], adg_pt->n_nfa);
      printf (",  ");
      print_iv (adg_pt->fixed_bit_masks[i], adg_pt->n_nfa);
      printf ("\n");
   }
   printf ("Nonfixed bit masks.\n");
   for (i = 0; i < adg_pt->n_non_fixed_bits; i++)
   {
      print_iv (adg_pt->non_fixed_bit_masks[i], adg_pt->n_nfa);
      printf ("\n");
   }

   printf ("Number of individuals with observed trait values: %d\n",
           adg_pt->n_obs);

   printf ("Indices to alleles of observed individuals:\n");
   printf ("Trait values of observed individuals:\n");

   for (i = 0; i < adg_pt->n_obs; i++)
   {
      printf ("(%2d, %2d) ", adg_pt->i_obsa[i * 2], adg_pt->i_obsa[i * 2 + 1]);
   }
   printf ("\n");

   for (i = 0; i < adg_pt->n_obs; i++)
   {
      printf ("%8.2e ", adg_pt->y[i]);
   }
   printf
      ("\n===========================================================================\n\n\n");
}

/**
 * Print a founder genome label vector
 *
 * @param fgl founder genome label vector
 *
 * @ l_fgl length of the vector
 */
void print_fgl (int *fgl, int l_fgl)
{
   int   i;
   for (i = 0; i < l_fgl; i++)
   {
      printf ("%d,", fgl[i]);
      if (i & 1)
         printf ("\n");
   }
   printf ("\n");
}

/**
 * Print the marginal test p-values to stdout.
 *
 * @param pars_pt marginal test parameters
 */
void print_miv_pars (miv_pars_t * pars_pt)
{
   int   j;
   int   h;

   printf ("%9s  %8s", "Test Pos.", "p-values");
   printf ("\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      printf ("%5d:   ", pars_pt->pos[j]);
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         printf ("%5.3f ", pars_pt->p[h][j]);
      }
      printf ("\n");
   }
}

/**
 * Write the marginal test results, test statistics and  p-values, to files in plain text format
 *
 * @param pars_pt marginal test parameters
 *
 * @param file_t name of test statistic file.
 *
 * @param file_p name of p-value file.
 *
 */
void printf_miv_pars (miv_pars_t * pars_pt, file_name_t file_t,
                      file_name_t file_p)
{
   FILE *file_out;
   int   j;
   int   h;

   /* Write p-values to file. */
   file_out = fopen (file_p, "w");
   if (file_out == NULL)
   {
      printf ("Cannot open file %s for writing. \n", file_p);
      exit (1);
   }

   /* Header of the output file. */
   fprintf (file_out, "%s\t%s", "test_pos", "test_map");

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      fprintf (file_out, "\tpval%d", h);
   }
   fprintf (file_out, "\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      /* Index and map position for test position j */
      fprintf (file_out, "%d\t%f", pars_pt->pos[j], pars_pt->map[j]);
      /* Test p-values */
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         fprintf (file_out, "\t%f", pars_pt->p[h][j]);
      }
      fprintf (file_out, "\n");
   }
   fclose (file_out);

   /* Write test statistic values to file. */
   file_out = fopen (file_t, "w");
   if (file_out == NULL)
   {
      printf ("Cannot open file %s for writing. \n", file_t);
      exit (1);
   }

   /* Header of the output file. */
   fprintf (file_out, "%s\t%s", "test_pos", "test_map");

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      fprintf (file_out, "\tt%d", h);
   }
   fprintf (file_out, "\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      /* Index and map position for test position j */
      fprintf (file_out, "%d\t%f", pars_pt->pos[j], pars_pt->map[j]);
      /* Test statistic values */
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         fprintf (file_out, "\t%f", pars_pt->t[h][j]);
      }
      fprintf (file_out, "\n");
   }
   fclose (file_out);
}

/**
 * Write the conditional test p-values, to stdout.
 *
 * @param pars_pt marginal test parameters
 *
 */
void print_civ_pars (civ_pars_t * pars_pt)
{
   int   j;
   int   h;

   printf ("%9s  %8s", "Test Pos.", "p-values");
   printf ("\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      printf ("%4d|%4d,%4d:   ", pars_pt->pos[j], pars_pt->posl[j],
              pars_pt->posr[j]);
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         printf ("%5.3f, ", pars_pt->p[h][j]);
      }
      printf ("\n");
   }
}

/**
 * Write the conditional test results, test statistics and p-values, to file.
 *
 * @param pars_pt marginal test parameters
 *
 * @param file_p name of test statistic file.
 *
 * @param file_p name of p-value file.
 *
 */
void printf_civ_pars (civ_pars_t * pars_pt, file_name_t file_t,
                      file_name_t file_p)
{
   FILE *file_out;
   int   j;
   int   h;

   /* Write p-values to file */
   file_out = fopen (file_p, "w");
   if (file_out == NULL)
   {
      printf ("Can not open file %s for writing.\n", file_p);
      exit (1);
   }
   fprintf (file_out, "%s\t%s\t%s\t%s\t%s\t%s",
            "test_pos", "test_pos_l", "test_pos_r",
            "test_map", "test_map_l", "test_map_r");
   for (h = 0; h < pars_pt->smpl_size; h++)
      fprintf (file_out, "\tpval%d", h);
   fprintf (file_out, "\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      fprintf (file_out, "%d\t%d\t%d\t", pars_pt->pos[j], pars_pt->posl[j],
               pars_pt->posr[j]);
      fprintf (file_out, "%f\t%f\t%f\t", pars_pt->map[j], pars_pt->mapl[j],
               pars_pt->mapr[j]);
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         fprintf (file_out, "\t%f", pars_pt->p[h][j]);
      }
      fprintf (file_out, "\n");
   }
   fclose (file_out);

   /* Write test statistic values to file */
   file_out = fopen (file_t, "w");
   if (file_out == NULL)
   {
      printf ("Can not open file %s for writing.\n", file_t);
      exit (1);
   }
   fprintf (file_out, "%s\t%s\t%s\t%s\t%s\t%s",
            "test_pos", "test_pos_l", "test_pos_r",
            "test_map", "test_map_l", "test_map_r");
   for (h = 0; h < pars_pt->smpl_size; h++)
      fprintf (file_out, "\tt%d", h);
   fprintf (file_out, "\n");

   for (j = 0; j < pars_pt->n_pos; j++)
   {
      fprintf (file_out, "%d\t%d\t%d\t", pars_pt->pos[j], pars_pt->posl[j],
               pars_pt->posr[j]);
      fprintf (file_out, "%f\t%f\t%f\t", pars_pt->map[j], pars_pt->mapl[j],
               pars_pt->mapr[j]);
      for (h = 0; h < pars_pt->smpl_size; h++)
      {
         fprintf (file_out, "\t%f", pars_pt->t[h][j]);
      }
      fprintf (file_out, "\n");
   }
   fclose (file_out);
}
