/***************************************************************************/
/*  Program to do a basic simulation and analysis of IBD inference with LD */
/*  Routines for forwards and back computation:  December 2008             */
/* Contains   actual_freq():  computes frequencies from real data          */
/*            make_freq():    puts frequencies into freq-array             */
/*            hap_set_penmat():  set penetrance matrices for analysis      */
/*            set_marg_ibd():  sets marginal ibd of states                 */
/*            set_transmat():  sets state transition matrix                */
/*            hap_fwd_compute():  does forward computation                 */
/*            hap_bck_compute():  does backward computation                */
/*            my_set_up():    sets up MORGAN framework                     */
/*                                                                         */
/***************************************************************************/

/* The avg_orbits subroutine is called to perform genotypic averaging.        */

void avg_orbits (double *input)
{
   double tmp;
   int   i, j;

   for (i = 0; i < num_orbits; ++i)
   {
      tmp = 0;
      for (j = 0; j < orbitSizes[i]; ++j)
         tmp += input[orbits[i][j]];
      tmp /= orbitSizes[i];

      for (j = 0; j < orbitSizes[i]; ++j)
         input[orbits[i][j]] = tmp;
   }

   return;
}

void reorder_tmat (void)
{
   static int order[] = { 0, 2, 3, 1, 4, 5, 8, 10, 9, 6, 12, 11, 13, 7, 14 };

   double **tmp_tmat;                   /* A temporary copy of tmat.          */

   int   i1, i2;                        /* Subscripts to use in for loops.    */
   int   o1, o2;                        /* Subscripts used to reorder the     */
   /*                                      transition rates.                  */

   /* Make temporary copy of the state transition matrix created by set_tmat. */

   tmp_tmat = alloc_dbl2d (num_states, num_states);
   memcpy (tmp_tmat[0], tmat[0], num_states * num_states * sizeof (double));

   /* Reorder the rows and columns in the 2009 state transition matrix to the */
   /* order used throughout the rest of ibd_haplo.  This order is detailed in */
   /* the static integer vector, order.                                       */

   for (i1 = 0; i1 < num_states; ++i1)
   {
      o1 = order[i1];
      for (i2 = 0; i2 < num_states; ++i2)
      {
         o2 = order[i2];
         tmat[i1][i2] = tmp_tmat[o1][o2];
      }
   }

   free_2d ((void *) &tmp_tmat);

   return;
}

void set_tmat_2009 (void)
{
   double hh, gg;                       /* Parameters of the Q-matrix.              */

   if (num_states != 15)
      give_up (ERA_stm, num_states);

   hh = 1 - PopKinship;
   gg = PopKinship;

   tmat = alloc_dbl2d (num_states, num_states);

   tmat[0][2] = tmat[0][3] = tmat[0][5] = tmat[0][6] = hh;
   tmat[1][4] = tmat[1][7] = hh;
   tmat[8][10] = tmat[8][13] = tmat[9][11] = tmat[9][12] = hh;
   tmat[0][0] = -4 * hh;
   tmat[1][1] = tmat[8][8] = tmat[9][9] = -2 * hh;

   tmat[2][0] = tmat[3][0] = tmat[5][0] = tmat[6][0] = 3 * gg;
   tmat[2][2] = tmat[3][3] = tmat[5][5] = tmat[6][6] = -3 * (gg + hh);
   tmat[2][4] = tmat[3][4] = tmat[5][7] = tmat[6][7] = hh;
   tmat[2][10] = tmat[2][12] = tmat[3][11] = tmat[3][13] = hh;
   tmat[5][10] = tmat[5][11] = tmat[6][12] = tmat[6][13] = hh;

   tmat[4][14] = tmat[7][14] = hh;
   tmat[4][1] = tmat[7][1] = gg;
   tmat[4][2] = tmat[4][3] = tmat[7][5] = tmat[7][6] = 2 * gg;
   tmat[4][4] = tmat[7][7] = -5 * gg - hh;

   tmat[10][14] = tmat[11][14] = tmat[12][14] = tmat[13][14] = hh;
   tmat[10][8] = tmat[11][9] = tmat[12][9] = tmat[13][8] = gg;
   tmat[10][2] = tmat[11][3] = tmat[12][2] = tmat[13][3] = 2 * gg;
   tmat[10][5] = tmat[11][5] = tmat[12][6] = tmat[13][6] = 2 * gg;
   tmat[10][10] = tmat[11][11] = tmat[12][12] = tmat[13][13] = -5 * gg - hh;

   tmat[14][10] = tmat[14][11] = tmat[14][12] = tmat[14][13] = gg;
   tmat[14][4] = tmat[14][7] = gg;
   tmat[14][14] = -6 * gg;

   /* Call reorder_tmat to reorder the 2009 state transition matrix to the    */
   /* order used throughout the rest of ibd_haplo.                            */

   reorder_tmat ();

   return;
}

void hap_set_penmat (double p0, double *ppvec, int num_haps, int ms, int phdat,
                     double errate, int igdx, int phased, int **statemat)
/*
 * p0 0 prob
 * ppvec output emit vector.
 * num_haps is the number of haplotypes in the current score set.
 * ms is the number of ibd states.  It is the length of ppvec.
 * phdat integer of data.
 * errate is the error rate.
 * igdx fyug  0 for phas 1 for genot 2 for pphas.
 * statemat is the IBD states matrix.
 */
{
   double allele_probs[2];
   double prob;

   unsigned int phdat_un;

   int   allele_count;
   int   class_count;
   int   num_classes;
   int   i, j, k;

   /* We will just deal with the data represented as an unsigned integer.     */
   /* If the data are longer than this (16 bits minimum), then there are too  */
   /* many IBD states to calculate anyway.                                    */

   phdat_un = (unsigned int) phdat;
   num_classes = 0;

   for (i = 0; i < ms; ++i)
   {
      for (j = 0; j < num_haps; ++j)
         num_classes = max (num_classes, statemat[i][j]);

      prob = 1.0;
      for (j = 0; j < num_classes; ++j)
      {
         allele_count = 0;              /* Count of 1 alleles observed.       */
         class_count = 0;               /* Count of haps in this IBD class.   */

         for (k = 0; k < num_haps; ++k)
         {
            if (statemat[i][k] == (j + 1)) /* Assume 1-indexed?               */
            {
               ++class_count;
               if (IS_SET (phdat_un, (num_haps - k - 1)))
                  ++allele_count;
            }
         }

         allele_probs[0] = p0 * pow (1 - errate, class_count - allele_count) *
            pow (errate, allele_count);
         allele_probs[1] = (1 - p0) * pow (errate, class_count - allele_count) *
            pow (1 - errate, allele_count);
         prob *= (allele_probs[0] + allele_probs[1]);
      }

      ppvec[i] = prob;
   }

   if ((PhaseofData == PARTIALPHASE && phased == 0) || PhaseofData == UNPHASED)
      avg_orbits (ppvec);               /* Then perform genotypic averaging.  */

   return;
}

void set_marg_ibd (int num_haps, int ms, double *ff, int **states)
{
   /* Function for computing the stationary distribution of the hidden Markov */
   /* process.  It is based on the version of Ewens' sampling formula given   */
   /* in Thompson, 2008, Theoretical Pop. Bio.                                */

   /* This is to rearrange the 4-chromosomes states so they match the legacy  */
   /* order (same as in transitions()).                                       */

   double bb;
   double lnormConst;

   int  *counts;
   int  *copyCounts;
   int   num_classes;
   int   i;
   int   j;

   counts = alloc_int1d (num_haps);
   copyCounts = alloc_int1d (num_haps);

   bb = PopKinship;

   /* Every class has this term in common.                                    */
   lnormConst = 0;
   for (i = 1; i < (num_haps - 1); ++i)
      lnormConst -= log (1 + i * bb);

   for (i = 0; i < ms; ++i)
   {
      /* Convert the 1,1,2,3 style states coding to counts.                   */
      /* copyCount[i] = number of classes which have (i + 1) copies in the    */
      /* state vector.  For example, if states[i] is [1,1,2,3] then           */
      /* copyCount[i] is [2,1,0,0].  That is there are 2 singly occurring     */
      /* haplotypes (2 and 3), 1 doubly occurring haplotypes (1) and no       */
      /* triply occurring or greater haplotypes.                              */
      /* counts is just the count of each symbol, an intermediate step.       */
      /* (Incidentally, these counts are the sufficient statistics for the    */
      /* ESF, since it is exchangeable).  CGG                                 */

      for (j = 0; j < num_haps; ++j)
      {
         counts[j] = 0;
         copyCounts[j] = 0;
      }

      /* Tabulate the counts of each symbol.                                  */
      for (j = 0; j < num_haps; ++j)
         ++counts[states[i][j] - 1];

      /* Tabulate copyCounts.                                                 */
      for (j = 0; j < num_haps; ++j)
         if (counts[j] > 0)
            ++copyCounts[counts[j] - 1];

      num_classes = 0;
      for (j = 0; j < num_haps; ++j)
         num_classes += copyCounts[j];

      ff[i] = lnormConst + (num_haps - num_classes) * log (bb)
         + (num_classes - 1) * log (1 - bb);

      for (j = 0; j < num_haps; ++j)
      {
         ff[i] += copyCounts[j] * lgamma ((double) (j + 1));

         /* There are a lot of terms here but most of them cancel;
          * the above line could be written as follows:
          *
          * ff[i] += log(factorial(num_haps));
          * ff[i] -= copyCounts[j]*log(j+1);
          * ff[i] -= log(factorial(copyCounts[j]));

          * ff[i] -= log(factorial(num_haps));
          * ff[i] += copyCounts[j]*log(factorial(j+1));
          * ff[i] += log(factorial(copyCounts[j]));
          *
          * The top 3 come from the rewritten ESF in Thompson 08 (see above).
          * They along with everything above give the probability of
          * the ORDERED class counts of the given IBD state.  We
          * must divide out the number of IBD states which share
          * this vector of ordered class counts.  This is
          * calculated from the copyCounts vector as follows.  Imagine the
          * class counts as a Young diagram of boxes.  There are n!
          * ways to place [1,...,n] in the boxes; we do not
          * distinguish between reorderings within classes/rows, so divide
          * by factorial(j)^copyCounts[j+1].  E.g. if we have two triples,
          * there are 3! * 3! indistinguishable rearrangements.
          * Additionally, we do not
          * distinguish between reorderings of classes of the same
          * size, so we divide by factorial(copyCounts[j]).  E.g. if
          * we have two triples, divide by 2! for the number of orders to
          * put the triples in.  One reference for this is Pitman
          * "Combinatorial stochastic processes", p. 16.  In general,
          * we are counting the number of set partitions
          * corresponding to a given integer partition.                       */
      }

      ff[i] = exp (ff[i]);
   }

   free_1d ((void *) &counts);
   free_1d ((void *) &copyCounts);

   return;
}

void set_transmat (int num_haps, int ns, double dd, double *ff, double **totmat)
/* Set the transition matrix at genetic distance dd, among ns states, at      */
/* equilibrium probabilities ff and change rate ffr between markers at        */
/* distance dd.  This is now called when needed within fwd/bk-compute.        */
{
   double ffr;

   int   i1;
   int   i2;

   /* Initialize totmat to all double precision 0.0's                         */
   memset (totmat[0], 0.0, ns * ns * sizeof (double));

   if (ns == 2)
   {
      ffr = 1.0 - exp (-dd * KinshipChanRat);
      for (i1 = 0; i1 < ns; ++i1)
         totmat[i1][i1] = (1.0 - ffr);

      for (i1 = 0; i1 < ns; ++i1)
         for (i2 = 0; i2 < ns; ++i2)
            totmat[i1][i2] += ffr * ff[i2];
   }
   else
   {
      for (i1 = 0; i1 < ns; ++i1)
      {
         ffr = exp (-dd * KinshipChanRat * tmat_rate[i1]);
         for (i2 = 0; i2 < ns; ++i2)
            totmat[i1][i2] = (1.0 - ffr) * tmat[i1][i2];
         totmat[i1][i1] += ffr;
      }
   }

   return;
}

void hap_fwd_compute (int *hd, int nd, int mph, int ms,
                      double **tmat2, double *allefrq,
                      double *fff, double *mdist, double **pprob, int igdx,
                      int *phasvec, int num_haps)
/* pprob[k, Sk] = P(Y0,...,Yk, Sk), 1 cluster, 2 states.                      */
/* NOTE now for 2-haps, 0 is IBD, 1 is NOT  */
/* NOTE missing data, marker positions etc are accommodated in hap_set_penmat.

 * fff: marginal probs.
 * pprob: matrix of forward probabilities.
 * mdist: vector of inter-locus distances.
 * pen2: penetrance vector (reset at each locus).
 * tmat2: transition matrix (reset at each locus).
 */
{
   double *pen2;
   double qn;

   int   phasint = 0;
   int   i, j, jj;

   pen2 = alloc_dbl1d (ms);

   printf (" Starting fwd-compute for %d markers\n", nd);
   fflush (stdout);

   /* If we are using the phase info, phasvec ought to be a non-null pointer. */
   if (igdx == 2)
      phasint = phasvec[0];

   hap_set_penmat (allefrq[0], pen2, num_haps, ms, hd[0], GenoErrRat, igdx,
                   phasint, state_matrix);

   for (j = 0; j < ms; ++j)
      pprob[0][j] = fff[j] * pen2[j];

   for (i = 1; i < nd; ++i)
   {
      qn = 0.0;

      if (igdx == 2)
         phasint = phasvec[i];
      hap_set_penmat (allefrq[i], pen2, num_haps, ms, hd[i], GenoErrRat, igdx,
                      phasint, state_matrix);
      set_transmat (num_haps, ms, mdist[i - 1], fff, tmat2);

      for (j = 0; j < ms; ++j)
      {
         pprob[i][j] = 0;
         for (jj = 0; jj < ms; ++jj)
            pprob[i][j] += pprob[i - 1][jj] * tmat2[jj][j] * pen2[j];
         qn += pprob[i][j];
      }
      for (j = 0; j < ms; ++j)
      {
         pprob[i][j] = pprob[i][j] / qn;
      }
   }

   free_1d ((void *) &pen2);

   return;
}

void hap_bck_compute (int *hd, int nd, int mph, int ms, double **qibd,
                      double **tmat2, double *allefrq,
                      double *fff, double *mdist, double **pprob, int igdx,
                      int *phasvec, int num_haps)
/*  pprob[k][Sk] = P(Yk,... Yn | Sk)  from the end down--replacing the fwd */
/*  qibd = P(ibd given all the data)
 *
 *  Same variable names as in hap_fwd_compute.                                */
{
   double *qi, qn;
   double *pen2;

   int   i, ii, j, jj;
   int   phasint = 0;

   pen2 = alloc_dbl1d (ms);
   qi = alloc_dbl1d (ms);

   printf (" Starting bck-compute for %d markers\n", nd);
   fflush (stdout);

   qn = 0.0;
   for (j = 0; j < ms; ++j)
      qn += pprob[nd - 1][j];
   /* when would qn ever not be == 1?  (normalized in hap_fwd_compute) */
   for (j = 0; j < ms; ++j)
      qibd[nd - 1][j] = pprob[nd - 1][j] / qn;

   ii = nd - 1;
   if (igdx == 2)
      phasint = phasvec[ii];
   hap_set_penmat (allefrq[ii], pen2, num_haps, ms, hd[nd - 1], GenoErrRat,
                   igdx, phasint, state_matrix);

   for (j = 0; j < ms; ++j)
      pprob[nd - 1][j] = pen2[j];

   for (i = nd - 2; i >= 0; --i)
   {
      set_transmat (num_haps, ms, mdist[i], fff, tmat2);
      /* we first calculate the marginal probability of state j at
       * marker i; the two pprob below are defined differently,
       * because we are replacing pprob[i][j] as we move backwards.
       * Each term is
       * Pr(S_i|Y_1:i)*Pr(S_i+1|S_i)*Pr(Y_i+1:n|S_i+1)
       * which works out to Pr(S_i==j|Y) after summing over S_i+1
       * and normalizing */
      for (j = 0; j < ms; ++j)
      {
         qi[j] = 0;
         for (jj = 0; jj < ms; ++jj)
            qi[j] += pprob[i][j] * tmat2[j][jj] * pprob[i + 1][jj];
      }
      qn = 0.0;
      for (j = 0; j < ms; ++j)
         qn += qi[j];
      for (j = 0; j < ms; ++j)
         qibd[i][j] = qi[j] / qn;

      for (j = 0; j < ms; ++j)
      {
         pprob[i][j] = 0.0;
         qn = 0.0;

         if (igdx == 2)
            phasint = phasvec[i];

         /* This is the actual backward probabilities calculation:            */
         /* Pr(Y_t|S_t) * Pr(S_t+1|S_t) * Pr(Y_i+1:n|S_t+1).                  */
         hap_set_penmat (allefrq[i], pen2, num_haps, ms, hd[i], GenoErrRat,
                         igdx, phasint, state_matrix);

         for (jj = 0; jj < ms; ++jj)
            pprob[i][j] += pen2[j] * tmat2[j][jj] * pprob[i + 1][jj];
         qn += pprob[i][j];
      }
      for (j = 0; j < ms; ++j)
         pprob[i][j] = pprob[i][j] / qn;
   }

   free_1d ((void *) &pen2);
   free_1d ((void *) &qi);

   return;
}

void hap_sample (int nd, int ms, int *simStates, double **tmat2,
                 double *fff, double *mdist, double **pprob, int igdx,
                 int num_haps)
/* backwards sampling of the entire sequence of IBD states
 * given marker data: the final forward probabilities are
 * Pr(state = S | Y = full data), so first sample this S.
 * Then given S_[j..n], Pr(S_[j-1] | S_[j..n],Y) =
 * Pr(S_[j-1] | S_j, Y_[1..j-1]) (by conditional independence)
 * proportional to Pr(S_j | S_[j-1])*Pr(S_[j-1]|Y_[1..j-1])
 * (by Bayes thm). The second factor is the forward prob for s
 * at locus j-1. */
{
   double *qi, qn, sampleThreshold, sampleSum;
   double *pen2;
   int   i, j;

   pen2 = alloc_dbl1d (ms);
   qi = alloc_dbl1d (ms);

   printf (" Starting back-simulation for %d markers\n", nd);
   fflush (stdout);

   sampleSum = 0;
   sampleThreshold = rans ();

   for (j = 0; j < ms; ++j)
   {
      sampleSum += pprob[nd - 1][j];
      if (sampleSum > sampleThreshold)
      {
         simStates[nd - 1] = j;
         break;
      }
   }

   for (i = nd - 2; i >= 0; --i)
   {
      set_transmat (num_haps, ms, mdist[i], fff, tmat2);
      for (j = 0; j < ms; ++j)
      {
         qi[j] = tmat2[j][simStates[i + 1]] * pprob[i][j];
      }
      qn = 0.0;
      for (j = 0; j < ms; ++j)
         qn += qi[j];
      for (j = 0; j < ms; ++j)
         qi[j] = qi[j] / qn;

      sampleSum = 0;
      sampleThreshold = rans ();
      for (j = 0; j < ms; ++j)
      {
         sampleSum += qi[j];
         if (sampleSum > sampleThreshold)
         {
            simStates[i] = j;
            break;
         }
      }
   }

   free_1d ((void *) &pen2);
   free_1d ((void *) &qi);

   return;
}
