/******************************************************************************/
/*                                                                            */
/* Program to do a basic analysis of IBD inference with no LD.                */
/*                                                                            */
/* It also includes options to simulate the ibd status of chromosomes.        */
/* We will make the chromosomes 2 Morgans, so most will have an IBD segment.  */
/*                                                                            */
/******************************************************************************/

#define DEFINE_DATA
#define PNAM "ibd_haplo"

#include "parsie.h"
#include "rans.h"
#include "markers.h"
#include "ibd_haplo.h"

#include "haplo_process.c"
#include "transitions.c"
#include "hap_fwd_bk_compute.c"
#include "orbits.c"

void  set_ProbIx (void);
void  init_ibd_haplo (void);

int main (int argc, char *argv[])
{
   /* When the data are partially phased, we can use Morgan's extra input     */
   /* file for reading in the partial phasing data.  This will be the only    */
   /* situation where the iXtra FILE variable will be needed.                 */

   FILE *iXtra;
   FILE *oScorF;

   FILTAB_PTR oXtra_pft;
   FILE *oXtra = NULL;

   double *snppos;                      /* Selected marker positions.         */
   double *allefreq;                    /* Allele frequencies at each marker. */

   double **qq_ibd_final;
   double **qq_ibd_order;
   double **pprob;
   double **qq_ibd;
   int  *simStates;

   char  temp_buff[12];                 /* A temporary buffer to format the   */
   /*                                      highest selected marker number and */
   /*                                      highest selected score set number  */
   /*                                      to.                                */

   /* Arrays of phasing data.                                                 */
   int **phasdatindiv;                  /* First is for individuals.          */
   int  *phasdat;                       /* Second for pairs of individuals.   */
   char **phas_nam;                     /* Names from phasing data file.      */

   int  *phasptr;

   int   nstmp;                         /* Temporary number of states.        */
   int   nph;                           /* Number of phenotypes.              */

   int **haplodat;                      /* Haplotypic data.                   */
   int  *hapdat;

   int   indiv;                         /* Subscript for current individual.  */

   int   tmp1, tmp2;                    /* The two tmp variables are holders  */
   /*                                      for swapping alleles.              */

   int   nc;                            /* Character length of longest name.  */
   int   nm;                            /* Character length of highest        */
   /*                                      numbered selected marker.          */
   int   nsn;                           /* Character length of highest        */
   /*                                      numbered score set.                */
   int   nst;                           /* Character length of largest number */
   /*                                      of possible IBD states.            */

   int   ss_size;                       /* Size of the current score set.     */
   int   ss_size_ndx;                   /* Current score set size subscript.  */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to score set.            */
   int   g1;                            /* Subscript to proband gamete.       */

   int   ix;                            /* Subscript for current individual.  */
   int   m1;                            /* Subscript to current marker.       */

   int   i1, i2;                        /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* The executable portion of the main program starts here.                    */
/*                                                                            */
/******************************************************************************/

   initMorgProg (argv);

   /* Parse input parameter statements.                                       */
   if (!proc_pars (argc, argv, PNAM))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* Check statements for completeness and consistency.                      */
   /* Set up global parameters and data structures.                           */
   if (!proc_auto ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   /* Initialize the integer variable, nc, to the number of characters in the */
   /* longest name of an observed individual in the dataset.                  */

   nc = 0;
   if (NObs > 0)
   {
      for (i1 = 0; i1 < NObs; ++i1)
      {
         i2 = strlen (ObsNam[i1]);
         nc = i2 > nc ? i2 : nc;
      }
   }

   /* To make the data written to the output scores file line up nicely, find */
   /* out how many characters are needed to write the marker number of the    */
   /* highest numbered selected marker.                                       */

   sprintf (temp_buff, "%d", MarSel[NMarks - 1]);
   nm = strlen (temp_buff);

   /* We will also need to know how many characters are needed to write the   */
   /* highest numbered proband gamete score set number to.                    */

   nsn = 0;
   for (s1 = 0; s1 < NScSet[c1]; ++s1)
   {
      sprintf (temp_buff, "%d", ScSetNo[c1][s1]);
      nsn = max (nsn, strlen (temp_buff));
   }

   /* Compute how many characters are needed to write the largest number of   */
   /* possible IBD states for any score set.  This will be for any score set  */
   /* with the most proband gametes.  For aesthetic reasons we don't want     */
   /* this to be less than two characters.                                    */

   sprintf (temp_buff, "%d", PTgg[MxGam]);
   nst = max (2, strlen (temp_buff));

   /* Call set_ProbIx to save the needed indices in ProbIx.                   */

   set_ProbIx ();

   if (PhaseofData == PARTIALPHASE)
   {
      /* Open the extra input file to read in the partial phasing data.       */

      if (!(iXtra = open_fil (XTRA)))
         give_up (ERR_qit);

      phas_nam = alloc_char2d (NObs, MAXNAMLEN + 1);
      phasdatindiv = alloc_int2d (NObs, NMarks);

      /* Read in one line of phasing data for each individual's pair of       */
      /* haplotypes.                                                          */

      for (ix = 0; ix < NObs; ++ix)
      {
         fscanf (iXtra, "%s %*s", phas_nam[ix]);
         if (strncmp (ObsNam[ix], phas_nam[ix], MAXNAMLEN) != 0)
         {
            fprintf (stderr, " A name, \"%s\", read from the", phas_nam[ix]);
            fprintf (stderr, " phasing data file does not match the");
            fprintf (stderr, " corresponding name, \"%s\", from", ObsNam[ix]);
            fprintf (stderr, " the marker data file  (W)\n\n");
         }

         for (m1 = 0; m1 < NMarks; ++m1)
            fscanf (iXtra, "%d", &(phasdatindiv[ix][m1]));
      }

      /* Close the extra input file since we are done reading in the partial  */
      /* phasing data.                                                        */

      fclose (iXtra);
   }

   /* Compute the positions of all the markers involved in the analysis.      */

   snppos = alloc_dbl1d (NMarks);
   snppos[0] = FirstMarkPos;
   for (m1 = 1; m1 < NMarks; ++m1)
      snppos[m1] = snppos[m1 - 1] + Mdist[m1 - 1];

   /* Copy the allele frequencies from the global variable, Qq, to the local  */
   /* variable, allefreq[].                                                   */

   allefreq = alloc_dbl1d (NMarks);
   for (m1 = 0; m1 < NMarks; ++m1)
      allefreq[m1] = Qq[m1 + 1][1];

   /* Open the output scores file.                                            */

   if (!(oScorF = open_fil (OSCORE)))
      give_up (ERR_qit);

   /* If the user has named an output extra file, open it now.                */

   oXtra_pft = get_filent (OXTR);
   if (oXtra_pft != NULL && oXtra_pft->fname)
      if (!(oXtra = open_fil (OXTR)))
         give_up (ERR_qit);

   fprintf (stdout, "\n");

   /* Call init_ibd_haplo to initialize those constructs specific to each     */
   /* score set size.  All of the pointers to these contructs are saved in    */
   /* score_set_size_pointers.                                                */

   init_ibd_haplo ();

   /* In the main loop in ibd_haplo compute the IBD probabilities for each    */
   /* score set indicated by the ibd_haplo user.                              */

   for (s1 = 0; s1 < NScSet[c1]; ++s1)  /* For each score set:                */
   {
      ss_size = NPrGam[c1][s1];         /* Get size of the current score set. */

      printf (" === Start of the computation of IBD probabilities for score");
      printf (" set %*d ===\n", nsn, ScSetNo[c1][s1]);
      fflush (stdout);

      haplodat = alloc_int2d (ss_size, NMarks);
      hapdat = alloc_int1d (NMarks);

      /* Copy the necessary genotypic data from the ObsMark[][][] array.      */

      for (g1 = 0; g1 < ss_size; ++g1)
      {
         ix = ProbIx[c1][s1][g1];

         for (m1 = 0; m1 < NMarks; ++m1)
            haplodat[g1][m1] = ObsMark[ix][m1][ProbGam[c1][s1][g1]];
      }

      /* Make the changes needed in haplodat to accomodate the partially      */
      /* phased data case.                                                    */

      phasptr = NULL;                   /* Default to no phasing data.        */

      if (PhaseofData == PARTIALPHASE)
      {
         phasdat = alloc_int1d (NMarks);

         for (indiv = 0; indiv < ss_size / 2; ++indiv)
         {
            ix = (s1 * 2) + indiv;

            for (m1 = 0; m1 < NMarks; ++m1)
            {
               if (phasdatindiv[ix][m1] == DISCORDANT)
               {
                  tmp1 = haplodat[2 * indiv][m1];
                  tmp2 = haplodat[2 * indiv + 1][m1];
                  haplodat[2 * indiv][m1] = tmp2;
                  haplodat[2 * indiv + 1][m1] = tmp1;
                  phasdatindiv[ix][m1] = CONCORDANT;
               }

               /* It's only phased for the HMM model if BOTH are phased.      */
               phasdat[m1] = phasdatindiv[2 * s1][m1] *
                  phasdatindiv[2 * s1 + 1][m1];
            }
         }

         if (DebugMain)
         {
            printf (" Individual name %*s:", nc, ObsNam[2 * s1]);
            for (m1 = 0; m1 < NMarks; ++m1)
               printf (" %d", phasdatindiv[2 * s1][m1]);
            printf ("\n");

            printf (" Individual name %*s:", nc, ObsNam[2 * s1 + 1]);
            for (m1 = 0; m1 < NMarks; ++m1)
               printf (" %d", phasdatindiv[2 * s1 + 1][m1]);
            printf ("\n");

            printf (" Pair (4-haplotypes) %*s %*s:", nc, ObsNam[2 * s1], nc,
                    ObsNam[2 * s1 + 1]);
            for (m1 = 0; m1 < NMarks; ++m1)
               printf (" %d", phasdat[m1]);
            printf ("\n");

            fflush (stdout);
         }

         /* When using phased data, set the relevant pointer.                 */
         phasptr = phasdat;
      }

      ss_size_ndx = ss_size - 1;        /* Set the score set size subscript.  */

      num_states = PTgg[ss_size];       /* Number of states for score set.    */

      /* Restore the data structures needed for this score set's size.        */

      state_matrix = score_set_size_pointers[ss_size_ndx].state_matrix;
      num_orbits = score_set_size_pointers[ss_size_ndx].num_orbits;
      orbitSizes = score_set_size_pointers[ss_size_ndx].orbitSizes;
      orbits = score_set_size_pointers[ss_size_ndx].orbits;
      marg_probs = score_set_size_pointers[ss_size_ndx].marg_probs;
      tmat_rate = score_set_size_pointers[ss_size_ndx].tmat_rate;
      tmat = score_set_size_pointers[ss_size_ndx].tmat;
      transmat = score_set_size_pointers[ss_size_ndx].transmat;

      /* The number of possible phenotypes is 2 ^ (number of haplotypes in    */
      /* the current score set.                                               */
      nph = 1 << ss_size;

      fprintf (stdout, " Markers = %d, proband gametes = %d,", NMarks, ss_size);
      fprintf (stdout, " states = %d, phenotypes = %d\n", num_states, nph);
      fflush (stdout);

      make_data (hapdat, haplodat, ss_size, 0, NMarks, 1, 0);

      pprob = alloc_dbl2d (NMarks, num_states);
      qq_ibd = alloc_dbl2d (NMarks, num_states);

      hap_fwd_compute (hapdat, NMarks, nph, num_states, transmat, allefreq,
                       marg_probs, Mdist, pprob, PhaseofData, phasptr, ss_size);

      if (!DebugGLDisp)
      {
         hap_bck_compute (hapdat, NMarks, nph, num_states, qq_ibd, transmat,
                          allefreq, marg_probs, Mdist, pprob, PhaseofData,
                          phasptr, ss_size);

         if (DebugLevel == 2)
         {
            fprintf (stderr, "\n");
            fprintf (stderr, " The qq_ibd matrix returned from");
            fprintf (stderr, " hap_bck_compute is:\n");

            for (i1 = 0; i1 < NMarks; ++i1)
            {
               fprintf (stderr, " [%4d] ", i1);
               for (i2 = 0; i2 < num_states; ++i2)
               {
                  fprintf (stderr, " %7.4f", qq_ibd[i1][i2]);
               }
               fprintf (stderr, "\n");
            }
         }
      }
      else
      {
         simStates = alloc_int1d (NMarks);
         hap_sample (NMarks, num_states, simStates, transmat, marg_probs, Mdist,
                     pprob, PhaseofData, ss_size);
      }

      /* If the user has named an output extra file, then output the score    */
      /* set identification information to the output extra file, 1 line per  */
      /* score set.                                                           */

      if (oXtra)
      {
         fprintf (oXtra, " %*d", nsn, ScSetNo[c1][s1]);
         fprintf (oXtra, " %2d", NPrGam[c1][s1]);
         fprintf (oXtra, " %*d", nst, PTgg[NPrGam[c1][s1]]);

         for (g1 = 0; g1 < ss_size; ++g1)
         {
            fprintf (oXtra, "  %*s", nc, ProbNam[c1][s1][g1]);
            fprintf (oXtra, " %d", ProbGam[c1][s1][g1]);
         }

         fprintf (oXtra, "\n");
         fflush (oXtra);
      }

      /* If the number of haplotypes in the current score set is 4, we still  */
      /* want to reorder the lexigraphically ordered qq_ibd[][] entries into  */
      /* the standard order previously used.                                  */

      if (ss_size == 4)
      {
         qq_ibd_order = alloc_dbl2d (NMarks, num_states);
         for (m1 = 0; m1 < NMarks; ++m1)
         {
            qq_ibd_order[m1][0] = qq_ibd[m1][0];
            qq_ibd_order[m1][1] = qq_ibd[m1][3];
            qq_ibd_order[m1][2] = qq_ibd[m1][1];
            qq_ibd_order[m1][3] = qq_ibd[m1][2];
            qq_ibd_order[m1][4] = qq_ibd[m1][4];
            qq_ibd_order[m1][5] = qq_ibd[m1][5];
            qq_ibd_order[m1][6] = qq_ibd[m1][9];
            qq_ibd_order[m1][7] = qq_ibd[m1][13];
            qq_ibd_order[m1][8] = qq_ibd[m1][6];
            qq_ibd_order[m1][9] = qq_ibd[m1][8];
            qq_ibd_order[m1][10] = qq_ibd[m1][7];
            qq_ibd_order[m1][11] = qq_ibd[m1][11];
            qq_ibd_order[m1][12] = qq_ibd[m1][10];
            qq_ibd_order[m1][13] = qq_ibd[m1][12];
            qq_ibd_order[m1][14] = qq_ibd[m1][14];
         }
         free_2d ((void *) &qq_ibd);
         qq_ibd = qq_ibd_order;
      }

      /* If partially/un- phased data, collapse OUTPUT of 15 states down to 9. */
      if (PhaseofData == PARTIALPHASE || PhaseofData == UNPHASED)
      {
         qq_ibd_final = alloc_dbl2d (NMarks, 9);
         for (m1 = 0; m1 < NMarks; ++m1)
         {
            qq_ibd_final[m1][0] = qq_ibd[m1][0];
            qq_ibd_final[m1][1] = qq_ibd[m1][1];
            qq_ibd_final[m1][2] = qq_ibd[m1][2] + qq_ibd[m1][3];
            qq_ibd_final[m1][3] = qq_ibd[m1][4];
            qq_ibd_final[m1][4] = qq_ibd[m1][5] + qq_ibd[m1][6];
            qq_ibd_final[m1][5] = qq_ibd[m1][7];
            qq_ibd_final[m1][6] = qq_ibd[m1][8] + qq_ibd[m1][9];
            qq_ibd_final[m1][7] = qq_ibd[m1][10] + qq_ibd[m1][11] +
               qq_ibd[m1][12] + qq_ibd[m1][13];
            qq_ibd_final[m1][8] = qq_ibd[m1][14];
         }
         free_2d ((void *) &qq_ibd);
         nstmp = 9;
      }
      else
      {
         qq_ibd_final = qq_ibd;
         nstmp = num_states;
      }

      if (!DebugGLDisp)
      {
         for (i1 = 0; i1 < NMarks; ++i1)
         {
            fprintf (oScorF, " %*d ", nsn, ScSetNo[c1][s1]);
            fprintf (oScorF, " %*d ", nm, MarSel[i1]);
            fprintf (oScorF, " %10.6f ", snppos[i1]);
            for (i2 = 0; i2 < nstmp; ++i2)
               fprintf (oScorF, " %7.4f", qq_ibd_final[i1][i2]);
            fprintf (oScorF, "\n");
         }
      }
      else
      {
         for (i1 = 0; i1 < NMarks; ++i1)
         {
            fprintf (oScorF, " %d", simStates[i1]);
         }
         fprintf (oScorF, "\n");
      }

      fprintf (stdout, "\n");
      if (s1 < NScSet[c1] - 1)          /* Unless this is the last score set, */
         fprintf (stdout, "\n");        /* print another blank line.          */
      fflush (stdout);

      if (PhaseofData == PARTIALPHASE)
         free_1d ((void *) &phasdat);

      if (qq_ibd_final)
         free_2d ((void *) &qq_ibd_final);

      free_2d ((void *) &haplodat);
      free_1d ((void *) &hapdat);
      free_2d ((void *) &pprob);
   }

   /* Close the scores file and output extra file.                            */

   fflush (oScorF);
   fclose (oScorF);

   if (oXtra)
   {
      fflush (oXtra);
      fclose (oXtra);
   }

   /* We are done, so all of the allocated storage can now be freed.          */

   free_1d ((void *) &snppos);
   free_1d ((void *) &allefreq);

   for (i1 = 0; i1 < MxGam; ++i1)
   {
      if (score_set_size_pointers[i1].state_matrix != NULL)
         free_2d ((void *) &score_set_size_pointers[i1].state_matrix);
      if (score_set_size_pointers[i1].orbitSizes != NULL)
         free_1d ((void *) &score_set_size_pointers[i1].orbitSizes);
      if (score_set_size_pointers[i1].orbits != NULL)
         free_2d ((void *) &score_set_size_pointers[i1].orbits);
      if (score_set_size_pointers[i1].marg_probs != NULL)
         free_1d ((void *) &score_set_size_pointers[i1].marg_probs);
      if (score_set_size_pointers[i1].tmat_rate != NULL)
         free_1d ((void *) &score_set_size_pointers[i1].tmat_rate);
      if (score_set_size_pointers[i1].tmat != NULL)
         free_2d ((void *) &score_set_size_pointers[i1].tmat);
      if (score_set_size_pointers[i1].transmat != NULL)
         free_2d ((void *) &score_set_size_pointers[i1].transmat);
   }
   free_1d ((void *) &score_set_size_pointers);

   if (PhaseofData == PARTIALPHASE)
   {
      free_2d ((void *) &phas_nam);
      free_2d ((void *) &phasdatindiv);
   }

   free_auto ();
   proc_free ();

   exit (EXIT_SUCCESS);
}

/******************************************************************************/
/*                                                                            */
/* This subroutine sets the values in the ProbIx array.  It stores the index  */
/* for each proband gamete to the named individual's marker data in ObsMark.  */
/* If a proband gamete name cannot be found in ObsNam, this subroutine prints */
/* a fatal error message before terminating the program.                      */
/*                                                                            */
/******************************************************************************/

void set_ProbIx (void)
{
   int   terminate;                     /* 0 -> All proband gamete names were */
   /*                                           found, so continue program.   */
   /*                                      1 -> One or more names not found,  */
   /*                                           so terminate the program.     */

   int   num_found;                     /* Number of named proband gametes    */
   /*                                      found in the marker data file.     */

   int   ind_ndx;                       /* Subscript for proband gamete.      */
   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to current score set.    */
   int   g1;                            /* Subscript to this proband gamete.  */
   int   ga;                            /* Subscript to alternate gamete.     */
   int   i1;                            /* Subscript to current individual.   */

   /* Start with the assumption that there won't be any names not found.      */
   terminate = FALSE;

   /* Initialize the number of named proband gametes found in the marker data */
   /* file to 0.                                                              */

   num_found = 0;

   /* We cannot continue executing code in the set_ProbIx subroutine without  */
   /* at least one "set proband gametes ..." parameter statement.             */

   if (NScSet == NULL)
      give_up (ERA_spg);

   for (s1 = 0; s1 < NScSet[c1]; ++s1)
   {
      for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1)
      {
         /* Initialize ind_ndx to INVALID_INDEX so that we can determine if   */
         /* the gamete's name was found among the names read in from the      */
         /* marker data file.                                                 */

         ind_ndx = INVALID_INDEX;

         for (i1 = 0; i1 < NObs; ++i1)
         {
            if (strncmp (ProbNam[c1][s1][g1], ObsNam[i1], MAXNAMLEN) == 0)
            {
               ProbIx[c1][s1][g1] = ind_ndx = i1;
               ++num_found;
               break;
            }
         }

         if (ind_ndx == INVALID_INDEX)
         {
            give_err (ERR_mpg, ProbNam[c1][s1][g1]);
            terminate = TRUE;
         }
      }

      /* If the data are not being analyzed as phased data we need to check   */
      /* that the proband gametes are only for sets of whole individuals.     */

      if (PhaseofData != PHASED)
      {
         for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1)
         {
            /* The following statement computes the subscript to the proband  */
            /* gamete which should be paired with the current proband gamete. */
            /* In other words, if g1 is even ga = g1 + 1, and if g1 is odd    */
            /* ga = g1 - 1.                                                   */

            ga = ((g1 / 2) * 2 == g1) ? g1 + 1 : g1 - 1;

            if (ga == NPrGam[c1][s1]
                || strcmp (ProbNam[c1][s1][g1], ProbNam[c1][s1][ga]))
            {
               give_err (ERR_nwi, ScSetNo[c1][s1], ProbNam[c1][s1][g1]);
               terminate = TRUE;
            }
         }
      }
   }

   if (num_found == 0)                  /* No named proband gametes found?    */
      give_up (ERA_npg);

   if (terminate)
      give_up (ERR_qit);

   return;
}

/******************************************************************************/
/*                                                                            */
/* This subroutine initializes those constructs which differ depending on the */
/* score set sizes needed to compute the IBD probabilities for the score sets */
/* the user has indicated in their Morgan parameter statements.               */
/*                                                                            */
/******************************************************************************/

void init_ibd_haplo (void)
{
   int  *printed;                       /* Vector used to remember whether or */
   /*                                      not output has already been done   */
   /*                                      for each score set.                */

   int   ss_size;                       /* Size of the current score set.     */
   int   ss_size_ndx;                   /* Current score set size subscript.  */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to score set.            */

   int   i1, i2;                        /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* The executable portion of this subroutine starts here.                     */
/*                                                                            */
/******************************************************************************/

   /* Construct those marginal IBD probabilities vectors and transition       */
   /* matrices which we will need for the proband gamete score sets the       */
   /* ibd_haplo user has indicated.                                           */

   score_set_size_pointers = malloc (MxGam * sizeof (SS_SIZE_STRUCTURE));
   alloc_chk (WHERE, score_set_size_pointers);
   memset (score_set_size_pointers, 0, MxGam * sizeof (SS_SIZE_STRUCTURE));

   for (s1 = 0; s1 < NScSet[c1]; ++s1)
   {
      ss_size = NPrGam[c1][s1];         /* Get size of the current score set. */
      ss_size_ndx = ss_size - 1;        /* Set the score set size subscript.  */

      if (score_set_size_pointers[ss_size_ndx].transmat == NULL)
      {
         /* So long as it has not previously been done, call the              */
         /* get_ibd_state_mat subroutine to obtain a matrix of IBD states for */
         /* the current score set size.  This subroutine also has to compute  */
         /* a vector of 0's and 1's reflecting where there are duplicate      */
         /* rows.  Since the set_tmat_2011 subroutine will also need this     */
         /* vector later, it will be retained as the duplicated_rows vector.  */

         get_ibd_state_mat (ss_size, &num_states, &duplicated_rows,
                            &state_matrix);
         score_set_size_pointers[ss_size_ndx].state_matrix = state_matrix;

         /* Construct the genotypic averaging orbits data matrix.             */

         orbits = alloc_int2d (num_states, num_states);
         orbitSizes = alloc_int1d (num_states);

         if (PhaseofData != PHASED)
            makeOrbits (ss_size, num_states, state_matrix);

         score_set_size_pointers[ss_size_ndx].num_orbits = num_orbits;
         score_set_size_pointers[ss_size_ndx].orbitSizes = orbitSizes;
         score_set_size_pointers[ss_size_ndx].orbits = orbits;

         /* Compute the vector of marginal ibd probabilities for the current  */
         /* score set size.                                                   */

         marg_probs = alloc_dbl1d (num_states);
         set_marg_ibd (ss_size, num_states, marg_probs, state_matrix);
         score_set_size_pointers[ss_size_ndx].marg_probs = marg_probs;

         if (num_states > 2)
         {
            /* Compute tmat for this score set's number of haplotypes.  The   */
            /* memory for tmat is allocated in the set_tmat_20xx subroutines. */

            if (StateTransMat == STM_2011)
               set_tmat_2011 (ss_size, PopKinship, &tmat);
            else
               set_tmat_2009 ();

            tmat_rate = alloc_dbl1d (num_states);

            for (i1 = 0; i1 < num_states; ++i1)
               tmat_rate[i1] = -tmat[i1][i1];

            for (i1 = 0; i1 < num_states; ++i1)
            {
               for (i2 = 0; i2 < num_states; ++i2)
                  tmat[i1][i2] = tmat[i1][i2] / tmat_rate[i1];
               tmat[i1][i1] = 0.0;
            }

            for (i1 = 0; i1 < num_states; ++i1)
            {
               for (i2 = 0; i2 < num_states; ++i2)
                  tmat[i1][i2] = (1.0 - TransMatNullFr) * tmat[i1][i2] +
                     TransMatNullFr * marg_probs[i2];
            }

            score_set_size_pointers[ss_size_ndx].tmat_rate = tmat_rate;
            score_set_size_pointers[ss_size_ndx].tmat = tmat;
         }

         free_1d ((void *) &duplicated_rows);

         /* Compute transmat for this score set's number of haplotypes.       */

         transmat = alloc_dbl2d (num_states, num_states);
         set_transmat (ss_size, num_states, 1.0, marg_probs, transmat);
         score_set_size_pointers[ss_size_ndx].transmat = transmat;
      }
   }

   if (PrintLevel >= 3)
   {
      /* Print the equilibrium state probabilities and the state transition   */
      /* matrix once for each score set size indicated by the "set proband    */
      /* gametes" parameter statements.                                       */

      printed = alloc_int1d (MxGam);    /* Initialized to all 0's.            */

      for (ss_size = 2; ss_size <= MxGam; ++ss_size)
      {
         ss_size_ndx = ss_size - 1;     /* Set the score set size subscript.  */
         num_states = PTgg[ss_size];    /* Number of states for score set.    */

         if (printed[ss_size_ndx] == FALSE)
         {
            marg_probs = score_set_size_pointers[ss_size_ndx].marg_probs;
            if (marg_probs != NULL)
            {
               printf (" The %d chromosome equilibrium state", ss_size);
               printf (" probabilities are:\n");
               for (i1 = 0; i1 < num_states; ++i1)
                  printf (" %5.3f", marg_probs[i1]);

               printf ("\n\n");
               fflush (stdout);
            }

            transmat = score_set_size_pointers[ss_size_ndx].transmat;
            if (transmat != NULL)
            {
               printf (" The %d chromosome state transition matrix is:",
                       ss_size);
               for (i1 = 0; i1 < num_states; ++i1)
               {
                  printf ("\n");
                  for (i2 = 0; i2 < num_states; ++i2)
                     printf (" %7.5f", transmat[i1][i2]);
               }

               printf ("\n\n\n");
               fflush (stdout);
            }

            printed[ss_size_ndx] = TRUE;
         }
      }

      free_1d ((void *) &printed);
   }

   return;
}
