/******************************************************************************/
/*                                                                            */
/*  Program name: gl_lods.c (Estimate or calculate LodScore by component,     */
/*                           using the results of gl_auto).                   */
/*                                                                            */
/*  EAT, April 2010.                                                          */
/*  SML, June  2012.                                                          */
/*                                                                            */
/*  Note that this program is modified from lm_linkage.c in Lodscore.         */
/*                                                                            */
/*  Also note it will not do MCMC, but instead will take gl_auto output to    */
/*    compute lod score contributions at the selected markers.                */
/*                                                                            */
/******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS

#include "lodscore.h"
#include "IBDgraph.h"

/******************************************************************************/
/*                                                                            */
/* Storage for all the lodscore_elements.  This is a number of IBDgraphs by   */
/* number of selected markers matrix, where each entry in the matrix is a     */
/* lodscore_element (LSCORE_ELEM).  Only one lodscore_element in each         */
/* equivalence class will actually be used for computing the contribution to  */
/* the log likelihood.  The others will not be used.                          */
/*                                                                            */
/******************************************************************************/

static LSCORE_ELEM **lses_storage;      /* Storage for all lodscore_elements. */

/******************************************************************************/
/*                                                                            */
/* Vector of equivalence structures.  This is a number of selected markers    */
/* long vector of marker_classes structures (MARK_CLASS).                     */
/*                                                                            */
/******************************************************************************/

static MARK_CLASS *equivs_vector;       /* Vector of equivalence structures.  */

/******************************************************************************/
/*                                                                            */
/* Assigned lodscore_element pointers matrix.  This is a number of IBDgraphs  */
/* by number of selected markers matrix, where each entry in the matrix is a  */
/* lodscore_element pointer (LSE_PTR).  The pointers in this matrix point to  */
/* the lodscore_elements in lses_storage.  It is by reassigning the pointers  */
/* in this matrix that a single lodscore_element will be used for all the     */
/* graphs in a single equivalence class.                                      */
/*                                                                            */
/******************************************************************************/

static LSE_PTR **p_lse_index;           /* Lodscore_element pointers matrix.  */

/******************************************************************************/
/*                                                                            */
/* Graph class subscripts matrix.  This is a number of IBDgraphs by number of */
/* selected markers matrix, where each entry in the matrix is a long integer, */
/* which is the class_index for the corresponding IBDgraph and selected       */
/* marker.                                                                    */
/*                                                                            */
/******************************************************************************/

static long **class_index;              /* Graph class subscripts matrix.     */

static char **DGL_names;
static int *DGL_index;

#include "gl_lods_rtnes.c"

void  initialize_lodscore_element (LSE_PTR);
void  prnt_gl_lods (double *);
void  gl_lods_free (int);
void  collapse_lodscore_elements (long, cmi_ptr);
void  do_single_range (long, long, long);
void  do_single_lse (long, long);
long  comp_mark_ndx (long);
double comp_loglike (long);

int main (int argc, char *argv[])
{
   FILTAB_PTR ptb;                      /* File table pointer of OSCORE file. */
   FILE *fp2;                           /* Processed DGL graphs input file.   */

   IBDGraphList *ibd_graphs;            /* Linked list of IBD graphs.         */
   IBDGraphEquivalences *ige;           /* IBDGraphEquivalences pointer.      */
   IBDGraph *graph;                     /* Pointer to current IBD graph.      */

   HashTableIterator *hti;              /* Working hash table iterator.       */
   HashObject *ho;                      /* Pointer to current hash object.    */
   cmi_ptr p_mi;                        /* Pointer to a MarkerInfo object.    */

   markertype mark_lower;               /* First marker in equivalence class. */
   markertype mark_upper;               /* Last marker in equivalence class.  */

   LSE_PTR p_lse;                       /* Working lodscore_element pointer.  */

   double *overall_lod;                 /* Overall vector of log likelihoods. */
   double base_loglike;                 /* Base-point log P(Y_T).             */

   double delta_prob;                   /* Change in probability.             */
   double probability;                  /* Current marker's probability.      */

   long  graph_num;                     /* Current IBDgraph number.           */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  mark_num;                      /* Current marker number.             */
   long  mark_ndx;                      /* Current marker subscript.          */

   long  num_classes;                   /* Number of classes.                 */
   long  num_graphs;                    /* Number of graphs.                  */

   long  c1;                            /* Current class subscript.           */
   long  g1;                            /* Current graph subscript.           */

   int   nsw;                           /* Number of switches.                */
   int   ped_ndx;                       /* Pedigree subscript into nghd[].    */

   int   nn;                            /* Number of individuals in DGL file. */
   int   c_dgl;                         /* Reduced number of individuals.     */
   int   n_dgl;                         /* Number of IBDgraphs in DGL file.   */

   /* ....................................................................... */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setSampling ();

   /* At this point we have:
    *    PedSize -- original meaning of number read from pedigree file:
    *       for whom the data (possibly missing) has been stored in nghds.
    *    NMarks which is the number of markers at which lod scores are to
    *       be evaluated.
    *    MarSel which is a list of these markers by index in gl_auto output:
    *      Note: gl_auto is modified so that both programs can use
    *       original input marker indexing.
    *      Note: if a marker is called for that is not in the gl_auto output
    *       reconstruction of FGL labels will result in the evaluation at
    *       the proximal preceding marker that is in the gl_auto output.
    *    NIt is the number of scored FGL graphs to be processed from the
    *       gl_auto output.
    *    Curently it is assumed that there is a SINGLE component pedigree
    *       in the gl_auto output, or that GLOBAL MCMC was used
    *         (the latter is generally NOT recommended).
    *   Modified March 2011: EAT; number of individuals in gl_auto file will
    *     be determined on read-in; it is no longer provided by PedSize.      */

   if (DebugLevel == 2)
   {
      /*  Check the requested markers.                                        */

      printf (" %d selected markers:\n", NMarks);
      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
         printf ("  %d", MarSel[mark_ndx]);

      printf ("\n\n");
      fflush (stdout);

      /*  Check the data file input.                                          */

      for (ped_ndx = 0; ped_ndx < pedsize; ++ped_ndx)
      {
         printf ("  %s  %d  ", nghd[ped_ndx].self, nghd[ped_ndx].obs);
         if (!nghd[ped_ndx].obs)
            printf ("  N/A\n");
         else
         {
            if (TraitTyp == QUANTITATIVE)
               printf (" %f\n", nghd[ped_ndx].q_dat[0]);
            if (TraitTyp == DISCRETE)
               printf (" %d\n", nghd[ped_ndx].d_dat[0]);
         }
      }

      printf ("\n\n");
      fflush (stdout);
   }

/* First we will reduce the DGL graphs to observed individuals.               */
/* They are input in "extra file" and will be output to "output scores".      */
/* We will then close output scores, and re-read it for computations.         */

/* We will reduce the DGL graph to the set of observed individuals, and       */
/* output it by nghd index, not by name.  (Note the index is 0-origin.)       */

/* Names in DGL but not in nghd will be dropped.                              */
/* Names in DGL and in nghd but with nghd.obs == 0 will be dropped.           */
/* Names NOT in DGL but in nghd, with ngdh.obs != 0 will be WARNED.           */

   printf (" nFGL from dummy pedigree input = %d\n", nFGL);

   nn = dgl_index ();                   /* Number of individuals in DGL file. */
   c_dgl = dgl_strip (&n_dgl, nn);

   /* Here is where we now reset nFGL.                                        */

   if (n_dgl < NIt)
   {
      printf ("\n %d graphs were requested, but only %d were given (W)",
              NIt, n_dgl);
      NIt = n_dgl;
   }

   printf ("\n Number of individuals in each DGL graph = %d", c_dgl);
   printf ("\n Number of individuals in nghd structure = %d", pedsize);
   printf ("\n Reset number of FGL = %d", nFGL);

   fflush (stdout);

/* The DGL file is now in a format that the IBDgraph software can use.        */
/* So we have a smaller set to run, with added weights.                       */

   ptb = get_filent (OSCORE);
   if (!(fp2 = fopen (ptb->fname, "r"))) /* This is the processed DGL graphs. */
      give_up (ERR_qit);
   printf ("\n Reopened reduced DGL graph file %s\n", ptb->fname);

/* Allocate penetrances, and set Normal consts for quantitative traits.       */

   if (TraitTyp == QUANTITATIVE)
      set_qtlpen (loci[0].ve);

   alloc_gen_pen (nFGL, NLoci);
   printf ("\n Have alloced gen_pen nFGL=%d, NLoci=%d\n\n", nFGL, NLoci);
   fflush (stdout);

/* Now compute the lod scores for each ibd_graph.                             */

   ibd_graphs = NewIBDGraphList ();     /* Initializing the graph list.       */
   createIBDGraphs (ptb->fname, ibd_graphs); /* Build the graphs.             */

/* Allocate the storage needed for the arrays required for the rest of this   */
/* program.  Note that the entries in these arrays are not initialized.       */

   lses_storage = (LSCORE_ELEM **) malloc (NIt * sizeof (LSCORE_ELEM *));
   alloc_chk (WHERE, lses_storage);

   lses_storage[0] =
      (LSCORE_ELEM *) malloc (NIt * NMarks * sizeof (LSCORE_ELEM));
   alloc_chk (WHERE, lses_storage[0]);

   p_lse_index = (LSE_PTR **) malloc (NIt * sizeof (LSE_PTR *));
   alloc_chk (WHERE, p_lse_index);

   p_lse_index[0] = (LSE_PTR *) malloc (NIt * NMarks * sizeof (LSE_PTR));
   alloc_chk (WHERE, p_lse_index[0]);

   class_index = (long **) malloc (NIt * sizeof (long *));
   alloc_chk (WHERE, class_index);

   class_index[0] = (long *) malloc (NIt * NMarks * sizeof (long));
   alloc_chk (WHERE, class_index[0]);

   for (graph_ndx = 1; graph_ndx < NIt; ++graph_ndx)
   {
      lses_storage[graph_ndx] = lses_storage[graph_ndx - 1] + NMarks;
      p_lse_index[graph_ndx] = p_lse_index[graph_ndx - 1] + NMarks;
      class_index[graph_ndx] = class_index[graph_ndx - 1] + NMarks;
   }

   equivs_vector = malloc (NMarks * sizeof (MARK_CLASS));
   alloc_chk (WHERE, equivs_vector);

/* Initialize the p_lse_index to point to the corresponding lodscore_element  */
/* within lses_storage.                                                       */

   for (graph_ndx = 0; graph_ndx < NIt; ++graph_ndx)
      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
         p_lse_index[graph_ndx][mark_ndx] = &lses_storage[graph_ndx][mark_ndx];

/* Determine equivalence_classes for each selected marker.                    */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      mark_num = MarSel[mark_ndx];
      equivs_vector[mark_ndx].mark_num = mark_num;

      ige = IBDGraphEquivalenceClassesAtMarker (ibd_graphs, mark_num);
      IBDGraphEquivalences_InplaceSort (ige);

      if (DebugLevel == 3)
      {
         printf ("\n Equivalence classes at marker %ld:\n\n", mark_num);
         IBDGraphEquivalences_Print (ige);
         printf ("\n");
      }

      equivs_vector[mark_ndx].num_classes = ige->n_classes;
      num_classes = ige->n_classes;

      equivs_vector[mark_ndx].num_graphs = malloc (num_classes * sizeof (long));
      alloc_chk (WHERE, equivs_vector[mark_ndx].num_graphs);

      equivs_vector[mark_ndx].p_lse = malloc (num_classes * sizeof (LSE_PTR));
      alloc_chk (WHERE, equivs_vector[mark_ndx].p_lse);

      /* For each class initialize and point to a single lodscore_element.    */
      /* Note that the actual IBDgraph number is not retained.  Instead it is */
      /* used to set the class_index matrix.                                  */

      for (c1 = 0; c1 < num_classes; ++c1)
      {
         graph_num = ige->classes[c1].graphs[0]->id;
         graph_ndx = graph_num - 1;

         p_lse = p_lse_index[graph_ndx][mark_ndx];
         equivs_vector[mark_ndx].p_lse[c1] = p_lse;
         initialize_lodscore_element (p_lse);

         ++p_lse->num_refs;
         class_index[graph_ndx][mark_ndx] = c1;

         /* For each remaining graph in this equivalence class point to the   */
         /* first graph's lodscore_element.                                   */

         num_graphs = ige->classes[c1].n_graphs;
         equivs_vector[mark_ndx].num_graphs[c1] = num_graphs;

         for (g1 = 1; g1 < num_graphs; ++g1)
         {
            graph_num = ige->classes[c1].graphs[g1]->id;
            graph_ndx = graph_num - 1;

            p_lse_index[graph_ndx][mark_ndx] = p_lse;
            ++p_lse->num_refs;
            class_index[graph_ndx][mark_ndx] = c1;
         }
      }

      O_DecRef (ige);
   }

   printf ("\n Lod scores for %d ibd graphs at markers:\n", NIt);
   printf ("       ");
   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      printf ("     marker-%d", MarSel[mark_ndx]);
   printf ("\n");
   fflush (stdout);

   for (graph_num = 1; graph_num <= NIt; ++graph_num)
   {
      fgl_set (fp2, c_dgl);             /* Set the founder genome labels.     */

      if (DebugMain)
      {
         fprintf (stderr, "\nFor IBD graph %ld,\n", graph_num);
         dbgXXGG ();
      }

      graph_ndx = graph_num - 1;
      graph = Igl_ViewItem (ibd_graphs, (size_t) (graph_ndx));

      if (graph->dirty)
         IBDGraph_Refresh (graph);

      /* Now, for the current graph, the lodscore_elements for markers within */
      /* a single equivalence marker interval are collapsed to a single       */
      /* lodscore_element.                                                    */

      hti = Hti_New (graph->graph_hashes);

      while (Hti_Next (&ho, hti))
      {
         p_mi = H_Mi (ho);
         if (p_mi != NULL)
            collapse_lodscore_elements (graph_num, p_mi);
      }

      /* Now that the pointer constructs are complete, log likehoods for each */
      /* lodscore element need to be calculated.  For each selected marker,   */
      /* call the do_single_lse subroutine to calculate the log likelihood    */
      /* for the associated lodscore_element.                                 */

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         mark_num = equivs_vector[mark_ndx].mark_num;

         if (DebugLevel >= 4)
         {
            IBDGraphInvariantRegion (&mark_lower, &mark_upper, graph, mark_num);
            if (mark_ndx == 0)
               printf ("\n IBD graph %4ld is invariant on:", graph_num);
            else
               printf ("                                ");

            PrintInterval (mark_lower, mark_upper);
            printf ("\n");
         }

         do_single_lse (graph_num, mark_num);
      }

      if (DebugLevel >= 4)
         printf ("\n");

      /* Print the contributions to log likelihood for each graph at each of  */
      /* the selected markers.                                                */

      printf ("  grph-%ld", graph_num);
      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         p_lse = p_lse_index[graph_ndx][mark_ndx];
         printf ("   %f", p_lse->loglike);
      }

      printf ("\n");
      fflush (stdout);

      Hti_Delete (hti);
   }

   O_DecRef (ibd_graphs);
   printf ("\n");

   /* If the user has provided each component's base log likelihood, gl_lods  */
   /* can compute overall lod scores for the selected markers.                */

   if (Base_loglikes != NULL)
   {
      /* Allocate a vector for keeping each selected marker's lod score.      */
      overall_lod = alloc_dbl1d (NMarks);

      /* For now only one component is supported by gl_lods.                  */
      base_loglike = Base_loglikes[0] * log (10.0);

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         probability = 0.0;             /* Reset total probability to 0.0.    */

         num_classes = equivs_vector[mark_ndx].num_classes;
         for (c1 = 0; c1 < num_classes; ++c1)
         {
            num_graphs = equivs_vector[mark_ndx].num_graphs[c1];
            p_lse = equivs_vector[mark_ndx].p_lse[c1];

            delta_prob = exp (p_lse->loglike - base_loglike);
            probability += (num_graphs * delta_prob / NIt);
         }
         overall_lod[mark_ndx] = log10 (probability);
      }

      prntLodSHead ("Overall lod scores found by gl_lods");
      prnt_gl_lods (overall_lod);

      free_1d ((void *) &overall_lod);
   }
   else
   {
      fprintf (stderr, "\n\n");
      fprintf (stderr, " gl_lods is not able to provide lod scores unless");
      fprintf (stderr, " the user provides a base log\n likelihood using");
      fprintf (stderr, " the \"set base log likelihood X1 ...\" parameter");
      fprintf (stderr, " statement (W)\n");
   }

   /* Free up storage used by gl_lods.                                        */

   gl_lods_free (nn);

   fclose (fp2);

   exit (EXIT_SUCCESS);
}

/******************************************************************************/
/*                                                                            */
/* Initialize a lodscore_element.                                             */
/*                                                                            */
/******************************************************************************/

void initialize_lodscore_element (LSE_PTR p_lse)
{
   p_lse->need_loglike = TRUE;
   p_lse->num_refs = 0;
   p_lse->loglike = 0.0;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The following subroutine contains code merged from the prntLodScores and   */
/* the prntLodSPos subroutines in the print_lods.c file, which is part of the */
/* Markers library.  This subroutine is specifically tailored for the gl_lods */
/* program.  This subroutine outputs marker names and sex specific positions  */
/* for the selected markers only; no traits are involved.  Eventually this    */
/* subroutine should probably be moved to the end of the print_lods.c file.   */
/*                                                                            */
/******************************************************************************/

void prnt_gl_lods (double *lod_scores)
{
   LOCPOS_PTR p_p;                      /* Current locus pointer.             */

   long  mark_ndx;                      /* Current marker subscript.          */
   long  loc_ndx;                       /* Current locus subscript.           */

   int   nCol1;                         /* Width of marker names column.      */

   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   p_p = LocPos;
   mark_ndx = 0;

   for (loc_ndx = 0; loc_ndx < NLocPos; ++loc_ndx)
   {
      if (p_p->locus_typ == MARK_LOC)
      {
         printf (" %*s ", nCol1, MarkNam[mark_ndx]);
         printf ("%11.3f%11.3f   ", p_p->pos[MAL], p_p->pos[FEM]);
         printf (" %11.4f\n", lod_scores[mark_ndx]);
         ++mark_ndx;
      }
      ++p_p;
   }

   printf ("\n");
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* This subroutine is called to free up storage allocated for use by the      */
/* gl_lods program.                                                           */
/*                                                                            */
/******************************************************************************/

void gl_lods_free (int num_indivs)
{
   long  mark_ndx;                      /* Current marker subscript.          */

   int   i1;                            /* Current individual subscript.      */

   free_2d ((void *) &lses_storage);
   free_2d ((void *) &p_lse_index);
   free_2d ((void *) &class_index);

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      free_1d ((void *) &equivs_vector[mark_ndx].num_graphs);
      free_1d ((void *) &equivs_vector[mark_ndx].p_lse);
   }

   free_1d ((void *) &equivs_vector);

   for (i1 = 0; i1 < num_indivs; ++i1)
      free_1d ((void *) &DGL_names[i1]);
   free_1d ((void *) &DGL_names);

   free_1d ((void *) &DGL_index);

   return;
}

/******************************************************************************/
/*                                                                            */
/* This subroutine collapses the multiple lodscore_elements already in use    */
/* for the ranges in the marker interval passed as the second argument to     */
/* just one single lodscore_element for each range.                           */
/*                                                                            */
/******************************************************************************/

void collapse_lodscore_elements (long graph_num, cmi_ptr p_mi)
{
   long  mark_num;                      /* First marker in current range.     */
   long  mark_end;                      /* Last marker in current range.      */

   int   r1;                            /* Current range number subscript.    */

   if (p_mi->num_array_ranges == 0)
   {
      mark_num = p_mi->r.start;
      mark_end = p_mi->r.end;
      do_single_range (graph_num, mark_num, mark_end);
   }
   else
   {
      for (r1 = 0; r1 < p_mi->num_array_ranges; ++r1)
      {
         mark_num = p_mi->range_list[r1].start;
         mark_end = p_mi->range_list[r1].end;
         do_single_range (graph_num, mark_num, mark_end);
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Collapse the lodscore_elements already in use for the range passed to this */
/* subroutine to the single lodscore_element in use for the first selected    */
/* marker within this range.                                                  */
/*                                                                            */
/******************************************************************************/

void do_single_range (long graph_num, long mark_num, long mark_end)
{
   LSE_PTR p_lse;                       /* Lodscore_element pointer.          */

   long  mark_ndx;                      /* Current marker subscript.          */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  class_ndx;                     /* Current class subscript.           */

   if (mark_end == MARKER_PLUS_INFTY)
      mark_end = MarSel[NMarks - 1] + 1;

   p_lse = NULL;

   while (mark_num < mark_end)
   {
      mark_ndx = comp_mark_ndx (mark_num);
      if (mark_ndx >= 0)
      {
         graph_ndx = graph_num - 1;
         if (p_lse == NULL)
         {
            /* We have found the first lodscore_element in this range.        */
            p_lse = p_lse_index[graph_ndx][mark_ndx];
         }
         else
         {
            /* Otherwise, the previously found lodscore_element should be     */
            /* used for this selected marker as well as the earlier one.      */
            --(p_lse_index[graph_ndx][mark_ndx]->num_refs);
            p_lse_index[graph_ndx][mark_ndx] = p_lse;
            ++p_lse->num_refs;
         }

         /* In either case save the pointer to this lodscore_element in       */
         /* equivs_vector[mark_ndx].p_lse[class_ndx].                         */

         class_ndx = class_index[graph_ndx][mark_ndx];
         equivs_vector[mark_ndx].p_lse[class_ndx] = p_lse;
         ++p_lse->num_refs;
      }
      ++mark_num;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Compute the log likelihood for a single lodscore_element.                  */
/*                                                                            */
/******************************************************************************/

void do_single_lse (long graph_num, long mark_num)
{
   LSE_PTR p_lse;                       /* Lodscore_element pointer.          */

   long  mark_ndx;                      /* Current marker subscript.          */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */

   mark_ndx = comp_mark_ndx (mark_num);
   if (mark_ndx >= 0)
   {
      graph_ndx = graph_num - 1;
      p_lse = p_lse_index[graph_ndx][mark_ndx];

      if (p_lse->need_loglike == TRUE)
      {
         p_lse->loglike = comp_loglike (mark_ndx);
         p_lse->need_loglike = FALSE;
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Compute the current marker subscript for the marker number passed as the   */
/* argument to this subroutine.  This subroutine returns a -1 if the passed   */
/* marker number is not a selected marker.                                    */
/*                                                                            */
/******************************************************************************/

long comp_mark_ndx (long mark_num)
{
   long  mark_ndx;                      /* Current subscript into MarSel.     */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      if (mark_num == MarSel[mark_ndx])
         return mark_ndx;
   }

   /* Return a -1 if the marker is not found in MarSel.                       */

   return -1;
}

/******************************************************************************/
/*                                                                            */
/* This is a short subroutine which serves as the gl_lods interface to the    */
/* gen_pen subroutine.  This subroutine returns the log likelihood at the     */
/* selected marker indicated by the one argument passed to this subroutine.   */
/*                                                                            */
/******************************************************************************/

double comp_loglike (long mark_ndx)
{
   double loglike;

   int  *fgls[2];
   int   err;

#if 0
/* Currently the fgl_set subroutine stores founder genome labels starting at  */
/* the 0'th subscript into the gg0 and gg1 vectors.  This is not consistent   */
/* with the use of these two vectors elsewhere in Morgan.  Elsewhere the      */
/* FGL's stored in gg0 and gg1 start at the m_origin'th subscript, leaving    */
/* room for the trait(s) in front of the markers.  If and when fgl_set is     */
/* changed to do this, then the following commands will be needed.            */

   int   m_origin;                      /* Subscript to the first marker.     */
   int   fgl_ndx;                       /* Subscript to current FGL's.        */

   m_origin = max (NTLocs, 1);
   fgl_ndx = m_origin + mark_ndx;
#endif

   fgls[0] = gg0[mark_ndx];
   fgls[1] = gg1[mark_ndx];

   loglike = log (gen_pen (0, fgls, loci + 0, nFGL, &err));

   return loglike;
}
