/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "markers.h"

/* set_chromosomes(), print_chromosome(), set_lpflag(), 
 * alloc_chromo(), alloc_lpflag(), free_chromo(), free_lpflag()        
 */
void set_chromosomes (int nl, double *map, int nh, int *l0, double *r0, int ix)
/* routine to set the array of chromosome recombinations and orders for
   simulations and likelihood evaluations;
   nl = number of marker loci, map = recombination fractions between;
   nh = number of alternative positions for disease -- 0-one is simulation one
   map[1]...map[nl-1] - recomb fracts between marker loci 
   map[nl]            - .5                                       MAJ 03/02
   ix = sex of map -- 0 = female, 1 = male
   .             (was 0 = avg, 1 = male, 2 = female)             MAJ 02/02
   (r0,l0) vector of pairs giving position of disease;
   l0 is marker locus to the left and recombination r0 is to that marker;
   note r0 <= r(l0).
   (if disease extreme left, l0=0, and r0 is recom to locus 1),

   Also to set flags for log-prob diffs, so more efficient.      EAT 12/00
   Moved to set_lpflag()
 */
{
   int   i, l;

   for (i = 0; i <= nh; ++i)            /* set chromosome i */
   {
      if (l0[i] == 0)
      {                                 /* for trait first on chromosome */
         chromo[i][0] = 0;
         xecomo[ix][i][0] = r0[i];
         revchro[i][0] = 0;
         for (l = 1; l <= nl; ++l)
         {
            chromo[i][l] = l;
            xecomo[ix][i][l] = map[l];
            revchro[i][l] = l;
         }
      }
      else if (l0[i] == nl)
      {                                 /* for trait last on chromosome */
         for (l = 1; l <= nl; ++l)
         {
            chromo[i][l - 1] = l;
            xecomo[ix][i][l - 1] = map[l];
            revchro[i][l] = l - 1;
         }
         chromo[i][nl] = 0;
         xecomo[ix][i][nl - 1] = r0[i];
         xecomo[ix][i][nl] = .5;
         revchro[i][0] = nl;
      }
      else
      {                                 /* for trait between marker loci 
                                           on chromosome */
         for (l = 0; l < l0[i] - 1; ++l)
         {
            chromo[i][l] = l + 1;
            xecomo[ix][i][l] = map[l + 1];
            revchro[i][l + 1] = l;
         }

         l = l0[i] - 1;
         chromo[i][l] = l + 1;
         xecomo[ix][i][l] = r0[i];
         revchro[i][l + 1] = l;

         ++l;
         chromo[i][l] = 0;
         revchro[i][0] = l;
         xecomo[ix][i][l] = (map[l] - r0[i]) / (1.0 - 2.0 * r0[i]);

         for (l = l0[i] + 1; l <= nl; ++l)
         {
            chromo[i][l] = l;
            xecomo[ix][i][l] = map[l];
            revchro[i][l] = l;
         }
      }
   }
}

void print_chromosome (int nl, int hh, int ix)
{
   int   j;

   for (j = 0; j < nl; ++j)
      printf ("%3d %6.3f", chromo[hh][j], xecomo[ix][hh][j]);
   printf ("%3d\n", chromo[hh][nl]);
   fflush (stdout);

   return;
}
void set_lpflag (int nl, int nh, int ix)
/*
 * Routine to set the sex-specific marker maps, and the trait locus 
 * position for jnt_auto, and lm_auto programs;                EAT Mar 00   
 *
 * set_maps() removed                                          MAJ Nov 00 
 *
 * Sets flags for log-prob diffs, so more efficient.           EAT DEC 00
 * 
 *
 * nl = number of marker loci
 * nh = number of alternative positions for disease -- 0-one is simulation one.
 * ix = sex of map;  0 = female, 1 = male                      MAJ 02/02
 *      (was 0 = avg, 1 = male, 2 = female)
 */
{
   int   i, jj, jr, jl, jsr, jsl;

   for (i = 1; i <= nh; ++i)            /* set chromosome i */
   {
      for (jj = 0; jj < nl; ++jj)
      {
         lpflag[ix][i][jj] = 0;
         jr = chromo[i][jj + 1];
         jsr = chromo[0][jj + 1];
         jl = chromo[i][jj];
         jsl = chromo[0][jj];
         if ((jr != jsr) || (jl != jsl) ||
             (xecomo[ix][i][jj] != xecomo[ix][0][jj]))
            lpflag[ix][i][jj] = 1;
      }
   }
   return;
}
void alloc_chromo (int nl, int nh)
{

/* mallocs int     chromo  [nh+1][nl+1]
 *         int     revchro [nh+1][nl+1]
 *         double  xecomo  [2][nh+1][nl+1]                       MAJ Feb 02 */

   int   i, j;

   chromo = malloc ((nh + 1) * sizeof (int *));
   alloc_chk (WHERE, chromo);
   revchro = malloc ((nh + 1) * sizeof (int *));
   alloc_chk (WHERE, revchro);
   for (j = 0; j <= nh; ++j)
   {
      chromo[j] = malloc ((nl + 1) * sizeof (int));
      alloc_chk (WHERE, chromo[j]);
      revchro[j] = malloc ((nl + 1) * sizeof (int));
      alloc_chk (WHERE, revchro[j]);
   }
   xecomo = malloc (2 * sizeof (double **));
   alloc_chk (WHERE, xecomo);
   for (i = 0; i < 2; ++i)
   {
      xecomo[i] = malloc ((nh + 1) * sizeof (double *));
      alloc_chk (WHERE, xecomo[i]);
      for (j = 0; j <= nh; ++j)
      {
         xecomo[i][j] = malloc ((nl + 1) * sizeof (double));
         alloc_chk (WHERE, xecomo[i][j]);
      }
   }
   return;
}

void alloc_lpflag (int nl, int nh)
{
/* mallocs lpflag [2][nh+1][nl]                                  MAJ Feb 02 */

   int   i, j;

   lpflag = malloc (2 * sizeof (int **));
   alloc_chk (WHERE, lpflag);
   for (i = 0; i < 2; ++i)
   {
      lpflag[i] = malloc ((nh + 1) * sizeof (int *));
      alloc_chk (WHERE, lpflag[i]);
      for (j = 0; j <= nh; ++j)
      {
         lpflag[i][j] = malloc (nl * sizeof (int));
         alloc_chk (WHERE, lpflag[i][j]);
      }
   }
   return;
}
void free_chromo (int nh)
{
   int   i, j;

   if (!chromo)
      return;
   for (i = 0; i < 2; ++i)
   {
      for (j = 0; j <= nh; ++j)
         free (xecomo[i][j]);
      free (xecomo[i]);
   }
   for (j = 0; j <= nh; ++j)
   {
      free (chromo[j]);
      free (revchro[j]);
   }
   free (xecomo);
   xecomo = NULL;
   free (chromo);
   chromo = NULL;
   free (revchro);
   revchro = NULL;

   return;
}

void free_lpflag (int nh)
{
   int   i, j;

   if (!lpflag)
      return;
   for (i = 0; i < 2; ++i)
   {
      for (j = 0; j <= nh; ++j)
         free (lpflag[i][j]);
      free (lpflag[i]);
   }
   free (lpflag);
   lpflag = NULL;

   return;
}
