/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#define PNAM "univar"

/**
 * @defgroup univar univar
 * @ingroup polyem
 * @{
 */

#define PEELING

#include "polyem.h"

/**************************************************************************
 * program to test basic em equations : case of fixed + random + error
 * case of up to 100-individual pedigree  (now expanded to 250, but not
 * recommended)

 * This version modified April 1990 to allow use of full G-inverse,
 * instead of just the G for observed individuals

 * This is a test program that permits all alternative methods;
 * 1) likelihood evaluation by peeling, by Thompson & Shaw, by brute force,
 *    and by gaussian elimination.
 * 2) EM equations considering either all pedigree individuals, or just
 *    observed,
 * 3) quadratic form to use either G-obs or G--all-inverse

 * Updated August 28, 1991, to use new library.
 * Updates May 1992, to fixed effects routines, and to variance
 *   component EM-routines for greater compatibility with bivar
 * Updated Aug 1996 - cosmetic changes were made to the output
 *   Updated May 1977 - likelihood evaluation by gaussion elimination (SCH)
 *
 * modified, Aug 1997, for Version 2.0 revisions:
 * use of parameter file for user-friendly specification of modelling
 * parameters;
 * use of dynamic storage for all major arrays, with memory allocated just
 * prior to use, and freed as soon as no longer needed.  (MAJ)
 *
 *************************************************************************/

extern double kinship (Ind *, Ind *);

void uv_tau_fit (double *e, double *evg)
/* routine for residual variance   */
{
   double temp;
   temp = SigErr * (Nq - SigErr * invtrc (evg));
   SigErr = (innerprod (e, e, Nq) + temp) / Nq;
}

void sig_old (double *evg, double *e, double **gg)
{
   double temp;

   /* there are two alternative forms of the EM-equations for additive
    * variance;
    * so SigAdd can use either form depending on "what is missing"
    */

   temp = matprod (e, gg, Nq) * SigAdd * SigAdd / (SigErr * SigErr);
   SigAdd = (temp + SigAdd * SigErr * invtrc (evg)) / Nq; /* old form */
}

void uv_sig_fit (double *evg, double *b, double **fullmat)
{
   double temp3;

   /* NOTE: there are two alternative forms of the EM-equations for
    * additive variance;
    * uv_sig_fit can use either form depending on "what is missing"
    */

   temp3 = matprod (b, fullmat, pedsize);
   SigAdd = (temp3 + SigAdd * (SigErr * invtrc (evg) + (pedsize - Nq)))
      / pedsize;
}

void uv_itermo (double *xx, double *b, double **fullmat)
/* Note within this routine "xx" are the residuals from any fixed effects;
 * i.e. "y" and NOT original data x
 *
 * (SCH) Modified 21 May 1997 to deal with looped pedigrees by reading
 * terms from inverse of G matrix
 */
{
   int   i, it, obs, kid, mum, dad, sp;
   double alpha, zz, t, d;
   Marlist *mptr;
   Indlist *head;

   alpha = SigErr / SigAdd;

   for (it = 0; it < ItMo; it++)
   {
      d = 0.0;
      for (i = pedsize - 1; i >= 0; --i)
      {
         zz = fullmat[i + 1][i + 1] * alpha;
         obs = qdex[i];
         if (obs)
            zz += 1.0;
         t = xx[obs];

         /* Pick up non-zero terms from fullmat (G inverse).
          * We can avoid searching whole matrix, because only the
          * off-diagonal terms referring to parents, spouse(s) and
          * offspring are non-zero
          */

         dad = nghd[i].pa->index;
         mum = nghd[i].ma->index;
         if (mum != INVALID_INDEX)
            t -= b[mum] * fullmat[i + 1][mum + 1] * alpha;
         if (dad != INVALID_INDEX)
            t -= b[dad] * fullmat[i + 1][dad + 1] * alpha;
         mptr = nghd[i].marriages;
         while (mptr)
         {
            sp = mptr->current->spouse->index;
            if (sp != INVALID_INDEX)
               t -= b[sp] * fullmat[i + 1][sp + 1] * alpha;
            head = mptr->current->kids;
            while (head)
            {
               kid = head->current->index;
               t -= b[kid] * fullmat[i + 1][kid + 1] * alpha;
               head = head->next;
            }
            mptr = mptr->next;
         }
         t /= zz;

         /* Compute difference in new breeding value to assess */
         /* convergence */
         zz = b[i] - t;
         d += zz * zz;

         b[i] = t;
      }
      d = d / (SigAdd * (double) pedsize);
      if (d <= Diff)
         break;
   }
}

void uv_breeding_values (double *y, double *a, double *b, double *e,
                         double **fullmat)
{
   int   i;

   uv_itermo (y, b, fullmat);
   for (i = 0; i < pedsize; ++i)
      a[qdex[i]] = b[i];
   for (i = 1; i <= Nq; ++i)
      e[i] = y[i] - a[i];
}

void uv_em_iter (int k, double *a, double *b, double *epsi,
                 double *y, double **genmat, double *evgen, double **fullmat)
{                                       /* EM iteration */
   /* iterate for expected breeding values */
   int   i;
   double temp;
   out_fixed (y, 0);
   uv_breeding_values (y, a, b, epsi, fullmat);

   /* now evaluate likelihood */
   if ((k / ItWt) * ItWt == k)
   {
      likval (k, evgen, y, a, epsi, genmat);

      lik_direct (evgen, y, genmat);
      if (peelseq)
         lik_peel (peelseq, y);
      lik_gauss (y);
   }
   /* now implement EM equations for new SigAdd, SigErr */
   uv_sig_fit (evgen, b - 1, fullmat);
   /* sig_old (evgen, epsi, genmat);    */
   uv_tau_fit (epsi, evgen);
   /* now implement EM equations for new mu, SigErr */
   if (NFix == 0)
   {
      out_fixed (y, 0);
      uv_breeding_values (y, a, b, epsi, fullmat);
      mu[0] = 0.0;
      for (i = 1; i <= Nq; ++i)
      {
         y[i] = (x[0][i] - a[i]);
         mu[0] += y[i];
      }
      mu[0] = mu[0] / Nq;
      Mufti[0][0][1] = mu[0];
      uv_tau_fit (epsi, evgen);
   }
   /* or for new fixed effects */
   else
   {
      for (i = 1; i <= NFix; ++i)
      {
         out_fixed (y, 0);
         uv_breeding_values (y, a, b, epsi, fullmat);
         out_bar_one (a, y, x[0], i, 0);
         temp = fixeffit (y, i, 0);
         mu[0] = temp;
         Mufti[0][0][1] = temp;
         uv_tau_fit (epsi, evgen);
      }
   }
}

void g_full (double **fmat)
{
   int   i, j;
   for (i = 0; i < pedsize; ++i)
      for (j = i; j < pedsize; ++j)
      {
         fmat[i + 1][j + 1] = 2.0 * kinship (&(nghd[i]), &(nghd[j]));
         fmat[j + 1][i + 1] = fmat[i + 1][j + 1];
      }
}

void g_set (double **genmat)
{
   int   i, j;
   for (i = 1; i <= Nq; ++i)
      for (j = i; j <= Nq; ++j)
      {
         genmat[i][j] = 2.0 * kinship (&(nghd[idex[i]]), &(nghd[idex[j]]));
         genmat[j][i] = genmat[i][j];
      }
}

void uv_setvals (double *z, double *b)
{
   int   i;

   /* mu[0] is set to initial value of overall mean */
   mu[0] = Mufti[0][0][1];
   /* adjust data values for initial fixed effects */
   out_fixed (z, 0);
   z[0] = 0.0;

   /* set initial breeding values */
   for (i = 0; i < pedsize; ++i)
      b[i] = 0.5 * z[qdex[i]];
   x[0][0] = 0.0;
   b[0] = 0.0;
}

void uv_alloc_vect (double **a, double **b, double **epsi, double **y)
{
   *a = malloc ((Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *a);
   *b = malloc (pedsize * sizeof (double));
   alloc_chk (WHERE, *b);
   *epsi = malloc ((Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *epsi);
   *y = malloc ((Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *y);
}

void uv_free_vect (double *a, double *b, double *epsi, double *y,
                   double **genmat, double **fullmat, double *evgen)
{
   free ((void *) a);
   a = NULL;
   free ((void *) b);
   b = NULL;
   free ((void *) epsi);
   epsi = NULL;
   free ((void *) y);
   y = NULL;
   free ((void *) genmat[0]);
   genmat[0] = NULL;
   free ((void *) genmat);
   genmat = NULL;
   free ((void *) fullmat[0]);
   fullmat[0] = NULL;
   free ((void *) fullmat);
   fullmat = NULL;
   free ((void *) evgen);
   evgen = NULL;

   return;
}

int main (int argc, char *argv[])
{
   int   i, j, k;
   double *a, *epsi, *b, *y;
   double *evgen;
   double cc;
   double **genmat;
   double **fullmat;
   char *peelnam;

   PCleanUp = clean_poly;
   atexit (clean_morg);

   setupPoly (argc, argv, PNAM);

   if (!check_pars (Nq, max_cpsize))
      exit (EXIT_FAILURE);

   /* This version of univar with explicit matrix cross-checks should */
   /* not be used for large pedigrees */

   /* set G-obs (i.e. genmat),  and G-all-inverse (i.e. fullmat ) */

   genmat = malloc ((Nq + 1) * sizeof (double *));
   alloc_chk (WHERE, genmat);
   genmat[0] = malloc ((Nq + 1) * (Nq + 1) * sizeof (double));
   alloc_chk (WHERE, genmat[0]);
   for (i = 1; i <= Nq; ++i)
      genmat[i] = genmat[i - 1] + Nq + 1;
   g_set (genmat);

   if (ChkGmt)
   {
      printf (" G-matrix for observed individuals (%d x %d):\n\n", Nq, Nq);
      for (i = 1; i <= Nq; ++i)
      {
         for (j = 1; j <= Nq; ++j)
         {
            print_val5 (genmat[i][j]);
            if (j % 10 == 0)
               printf ("\n");
         }
         printf ("\n");
      }
      printf ("\n\n");
      fflush (stdout);
      exit (EXIT_SUCCESS);
   }

   /* eigenvalues are computed by cheigen  */

   evgen = malloc ((Nq + 1) * sizeof (double));
   alloc_chk (WHERE, evgen);

   cheigen (genmat, evgen, Nq);

   printf (" Eigenvalues have been computed\n\n");

   /* For testing, print eigenvalues and exit */

   if (ChkEig)
   {
      printf (" eigenvalues of G-observed:\n\n");
      for (i = 1; i <= Nq; ++i)
      {
         print_val5 (evgen[i]);
         if (i % 10 == 0)
            printf ("\n");
      }
      printf ("\n\n");
      fflush (stdout);
      exit (EXIT_SUCCESS);
   }
   if (ChkTrc)
   {
      cc = 0.0;
      for (i = 1; i <= Nq; ++i)
         cc += log (evgen[i]);
      printf (" check log-determinant of G =");
      print_val (cc);

      cc = 0.0;
      for (i = 1; i <= Nq; ++i)
         cc += 1.0 / evgen[i];
      printf ("\n check trace of G-inverse =  ");
      print_val (cc);
      printf ("\n\n");
      fflush (stdout);
   }
   fullmat = malloc ((pedsize + 1) * sizeof (double *));
   alloc_chk (WHERE, fullmat);
   fullmat[0] = malloc ((pedsize + 1) * (pedsize + 1) * sizeof (double));
   alloc_chk (WHERE, fullmat[0]);
   for (i = 1; i <= pedsize; ++i)
      fullmat[i] = fullmat[i - 1] + pedsize + 1;
   g_full (fullmat);
   invert_mat (fullmat, pedsize);

   /* allocate peeling variables  */

   peelseq = 0;

   peelnam = get_filent (PEEL)->fname;
   if (peelnam)
   {
      peelseq_set (peelnam);
      free ((void *) peelnam);
      alloc_peel (pedsize, 3);
   }

   /* allocate space for vectors  */

   uv_alloc_vect (&a, &b, &epsi, &y);

   y[0] = 0.0;

   /* set initial values */

   setfixeff ();

   uv_setvals (y, b);

   /* now ready for EM iteration  */

   for (k = 0; k <= ItEM; ++k)
      uv_em_iter (k, a, b, epsi, y, genmat, evgen, fullmat);

   out_fixed (y, 0);
   uv_breeding_values (y, a, b, epsi, fullmat);
   likval (k, evgen, y, a, epsi, genmat);

   free_input_data ();
   free_nghds ();
   uv_free_vect (a, b, epsi, y, genmat, fullmat, evgen);
   free_gauss ();
   if (peelseq)
      free_peel ();

   exit (EXIT_SUCCESS);
}

/** @} */
